from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    execute_pipeline,
)

from dagster.core.test_utils import single_output_transform


def _set_key_value(ddict, key, value):
    ddict[key] = value
    return value


def test_execute_solid_with_dep_only_inputs_no_api():
    did_run_dict = {}

    step_one_solid = single_output_transform(
        name='step_one_solid',
        inputs=[],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_one', True),
        output=OutputDefinition(),
    )

    step_two_solid = single_output_transform(
        name='step_two_solid',
        inputs=[InputDefinition('step_one_solid')],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_two', True),
        output=OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solids=[step_one_solid, step_two_solid],
        dependencies={
            'step_two_solid': {
                'step_one_solid': DependencyDefinition('step_one_solid')
            },
        },
    )

    # from dagster.utils import logging

    pipeline_result = execute_pipeline(pipeline)

    assert pipeline_result.success

    for result in pipeline_result.result_list:
        assert result.success

    assert did_run_dict['step_one'] is True
    assert did_run_dict['step_two'] is True


def test_execute_solid_with_dep_only_inputs_with_api():
    did_run_dict = {}

    step_one_solid = single_output_transform(
        name='step_one_solid',
        inputs=[],
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_one', True),
        output=OutputDefinition(),
    )

    step_two_solid = single_output_transform(
        name='step_two_solid',
        transform_fn=lambda context, args: _set_key_value(did_run_dict, 'step_two', True),
        inputs=[InputDefinition(step_one_solid.name)],
        output=OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solids=[step_one_solid, step_two_solid],
        dependencies={
            'step_two_solid': {
                step_one_solid.name: DependencyDefinition(step_one_solid.name)
            }
        },
    )

    pipeline_result = execute_pipeline(pipeline)

    for result in pipeline_result.result_list:
        assert result.success

    assert did_run_dict['step_one'] is True
    assert did_run_dict['step_two'] is True



from dagster import (
    OutputDefinition,
    Result,
    SolidDefinition,
    check,
)


def define_stub_solid(name, value):
    check.str_param(name, 'name')

    def _value_t_fn(*_args):
        yield Result(value)

    return SolidDefinition(
        name=name,
        inputs=[],
        outputs=[OutputDefinition()],
        transform_fn=_value_t_fn,
    )



import inspect
import os
import yaml

from dagster import check


def script_relative_path(file_path):
    '''
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test
    '''
    # from http://bit.ly/2snyC6s

    check.str_param(file_path, 'file_path')
    scriptdir = inspect.stack()[1][1]
    return os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path)


def load_yaml_from_path(path):
    check.str_param(path, 'path')
    with open(path, 'r') as ff:
        return yaml.load(ff)



from collections import namedtuple

import pytest
from dagster import (
    ConfigDefinition,
    types,
    Field,
    DagsterEvaluateValueError,
)


def test_noop_config():
    assert ConfigDefinition(types.Any)


def test_int_field():
    config_def = ConfigDefinition.config_dict(
        'SingleRequiredInt',
        {
            'int_field': Field(types.Int),
        },
    )

    assert config_def.config_type.evaluate_value({'int_field': 1}) == {'int_field': 1}


def test_int_fails():
    config_def = ConfigDefinition.config_dict(
        'SingleRequiredInt', {
            'int_field': Field(types.Int),
        }
    )

    with pytest.raises(DagsterEvaluateValueError):
        config_def.config_type.evaluate_value({'int_field': 'fjkdj'})

    with pytest.raises(DagsterEvaluateValueError):
        config_def.config_type.evaluate_value({'int_field': True})


def test_default_arg():
    config_def = ConfigDefinition.config_dict(
        'TestDefaultArg', {
            'int_field': Field(types.Int, default_value=2, is_optional=True),
        }
    )

    assert config_def.config_type.evaluate_value({}) == {'int_field': 2}


def _single_required_string_config_dict():
    return ConfigDefinition.config_dict(
        'SingleRequiredField', {'string_field': Field(types.String)}
    )


def _multiple_required_fields_config_dict():
    return ConfigDefinition.config_dict(
        'MultipleRequiredFields', {
            'field_one': Field(types.String),
            'field_two': Field(types.String),
        }
    )


def _single_optional_string_config_dict():
    return ConfigDefinition.config_dict(
        'SingleOptionalString', {'optional_field': Field(types.String, is_optional=True)}
    )


def _single_optional_string_field_config_dict_with_default():
    optional_field_def = Field(
        types.String,
        is_optional=True,
        default_value='some_default',
    )
    return ConfigDefinition.config_dict(
        'SingleOptionalStringWithDefault',
        {'optional_field': optional_field_def},
    )


def _mixed_required_optional_string_config_dict_with_default():
    return ConfigDefinition.config_dict(
        'MixedRequired', {
            'optional_arg': Field(
                types.String,
                is_optional=True,
                default_value='some_default',
            ),
            'required_arg': Field(types.String, is_optional=False),
            'optional_arg_no_default': Field(types.String, is_optional=True),
        }
    )


def _validate(config_def, value):
    return config_def.config_type.evaluate_value(value)


def test_single_required_string_field_config_type():
    assert _validate(_single_required_string_config_dict(), {'string_field': 'value'}) == {
        'string_field': 'value'
    }
    assert _validate(_single_required_string_config_dict(), {'string_field': None}) == {
        'string_field': None
    }

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_required_string_config_dict(), {})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_required_string_config_dict(), {'extra': 'yup'})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_required_string_config_dict(), {'string_field': 'yupup', 'extra': 'yup'})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_required_string_config_dict(), {'string_field': 1})


def test_multiple_required_fields_passing():
    assert _validate(
        _multiple_required_fields_config_dict(),
        {
            'field_one': 'value_one',
            'field_two': 'value_two',
        },
    ) == {
        'field_one': 'value_one',
        'field_two': 'value_two',
    }

    assert _validate(
        _multiple_required_fields_config_dict(),
        {
            'field_one': 'value_one',
            'field_two': None,
        },
    ) == {
        'field_one': 'value_one',
        'field_two': None,
    }


def test_multiple_required_fields_failing():
    with pytest.raises(DagsterEvaluateValueError):
        _validate(_multiple_required_fields_config_dict(), {})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_multiple_required_fields_config_dict(), {'field_one': 'yup'})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_multiple_required_fields_config_dict(), {'field_one': 'yup', 'extra': 'yup'})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(
            _multiple_required_fields_config_dict(), {
                'field_one': 'value_one',
                'field_two': 2
            }
        )


def test_single_optional_field_passing():
    assert _validate(_single_optional_string_config_dict(), {'optional_field': 'value'}) == {
        'optional_field': 'value'
    }
    assert _validate(_single_optional_string_config_dict(), {}) == {}

    assert _validate(_single_optional_string_config_dict(), {'optional_field': None}) == {
        'optional_field': None
    }


def test_single_optional_field_failing():
    with pytest.raises(DagsterEvaluateValueError):

        _validate(_single_optional_string_config_dict(), {'optional_field': 1})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_optional_string_config_dict(), {'dlkjfalksdjflksaj': 1})


def test_single_optional_field_passing_with_default():
    assert _validate(_single_optional_string_field_config_dict_with_default(), {}) == {
        'optional_field': 'some_default'
    }

    assert _validate(
        _single_optional_string_field_config_dict_with_default(), {'optional_field': 'override'}
    ) == {
        'optional_field': 'override'
    }


def test_mixed_args_passing():
    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(), {
            'optional_arg': 'value_one',
            'required_arg': 'value_two',
        }
    ) == {
        'optional_arg': 'value_one',
        'required_arg': 'value_two',
    }

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(), {
            'required_arg': 'value_two',
        }
    ) == {
        'optional_arg': 'some_default',
        'required_arg': 'value_two',
    }

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(), {
            'required_arg': 'value_two',
            'optional_arg_no_default': 'value_three',
        }
    ) == {
        'optional_arg': 'some_default',
        'required_arg': 'value_two',
        'optional_arg_no_default': 'value_three',
    }


def _single_nested_config():
    return ConfigDefinition(
        config_type=types.ConfigDictionary(
            'ParentType', {
                'nested':
                Field(
                    dagster_type=types.ConfigDictionary(
                        'NestedType',
                        {'int_field': Field(types.Int)},
                    )
                ),
            }
        )
    )


def _nested_optional_config_with_default():
    return ConfigDefinition(
        config_type=types.ConfigDictionary(
            'ParentType', {
                'nested':
                Field(
                    dagster_type=types.ConfigDictionary(
                        'NestedType',
                        {'int_field': Field(
                            types.Int,
                            is_optional=True,
                            default_value=3,
                        )}
                    )
                ),
            }
        )
    )


def _nested_optional_config_with_no_default():
    nested_type = types.ConfigDictionary(
        'NestedType',
        {
            'int_field': Field(
                types.Int,
                is_optional=True,
            ),
        },
    )
    return ConfigDefinition(
        config_type=types.ConfigDictionary(
            'ParentType',
            {'nested': Field(dagster_type=nested_type)},
        )
    )


def test_single_nested_config():
    assert _validate(_single_nested_config(), {'nested': {
        'int_field': 2
    }}) == {
        'nested': {
            'int_field': 2
        }
    }


def test_single_nested_config_failures():
    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_nested_config(), {'nested': 'dkjfdk'})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_nested_config(), {'nested': {'int_field': 'dkjfdk'}})

    with pytest.raises(DagsterEvaluateValueError):
        _validate(_single_nested_config(), {'nested': {'int_field': {'too_nested': 'dkjfdk'}}})


def test_nested_optional_with_default():
    assert _validate(_nested_optional_config_with_default(), {'nested': {
        'int_field': 2
    }}) == {
        'nested': {
            'int_field': 2
        }
    }

    assert _validate(_nested_optional_config_with_default(), {'nested': {}}) == {
        'nested': {
            'int_field': 3
        }
    }


def test_nested_optional_with_no_default():
    assert _validate(_nested_optional_config_with_no_default(), {'nested': {
        'int_field': 2
    }}) == {
        'nested': {
            'int_field': 2
        }
    }

    assert _validate(_nested_optional_config_with_no_default(), {'nested': {}}) == {'nested': {}}


CustomStructConfig = namedtuple('CustomStructConfig', 'foo bar')


class CustomStructConfigType(types.DagsterCompositeType):
    def __init__(self):
        super(CustomStructConfigType, self).__init__(
            'CustomStructConfigType',
            {
                'foo': Field(types.String),
                'bar': Field(types.Int),
            },
            lambda val: CustomStructConfig(foo=val['foo'], bar=val['bar']),
        )


def test_custom_composite_type():
    config_type = CustomStructConfigType()

    assert config_type.evaluate_value({
        'foo': 'some_string',
        'bar': 2
    }) == CustomStructConfig(
        foo='some_string', bar=2
    )

    with pytest.raises(DagsterEvaluateValueError):
        assert config_type.evaluate_value({
            'foo': 'some_string',
        })

    with pytest.raises(DagsterEvaluateValueError):
        assert config_type.evaluate_value({
            'bar': 'some_string',
        })

    with pytest.raises(DagsterEvaluateValueError):
        assert config_type.evaluate_value({
            'foo': 'some_string',
            'bar': 'not_an_int',
        })



import dagster

from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    Result,
    SolidDefinition,
    check,
    config,
    execute_pipeline,
    types,
)

from dagster.core.test_utils import single_output_transform

SingleValueDict = types.ConfigDictionary('SingleValueDict', {'value': Field(types.String)})


def define_pass_value_solid(name, description=None):
    check.str_param(name, 'name')
    check.opt_str_param(description, 'description')

    def _value_t_fn(info, _inputs):
        yield Result(info.config['value'])

    return SolidDefinition(
        name=name,
        description=description,
        inputs=[],
        outputs=[OutputDefinition(types.String)],
        config_def=ConfigDefinition(SingleValueDict),
        transform_fn=_value_t_fn,
    )


def test_execute_solid_with_input_same_name():
    a_thing_solid = single_output_transform(
        'a_thing',
        inputs=[InputDefinition(name='a_thing')],
        transform_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output=dagster.OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solids=[define_pass_value_solid('pass_value'), a_thing_solid],
        dependencies={'a_thing': {
            'a_thing': DependencyDefinition('pass_value')
        }},
    )

    result = execute_pipeline(
        pipeline,
        config.Environment(solids={'pass_value': config.Solid({
            'value': 'foo'
        })}),
    )

    assert result.result_for_solid('a_thing').transformed_value() == 'foofoo'


def test_execute_two_solids_with_same_input_name():
    input_def = InputDefinition(name='a_thing')

    solid_one = single_output_transform(
        'solid_one',
        inputs=[input_def],
        transform_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output=dagster.OutputDefinition(),
    )

    solid_two = single_output_transform(
        'solid_two',
        inputs=[input_def],
        transform_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(
        solids=[
            define_pass_value_solid('pass_to_one'),
            define_pass_value_solid('pass_to_two'),
            solid_one,
            solid_two,
        ],
        dependencies={
            'solid_one': {
                'a_thing': DependencyDefinition('pass_to_one')
            },
            'solid_two': {
                'a_thing': DependencyDefinition('pass_to_two')
            }
        }
    )

    result = execute_pipeline(
        pipeline,
        environment=config.Environment(
            solids={
                'pass_to_one': config.Solid({
                    'value': 'foo'
                }),
                'pass_to_two': config.Solid({
                    'value': 'bar'
                }),
            }
        )
    )

    assert result.success

    assert result.result_for_solid('solid_one').transformed_value() == 'foofoo'
    assert result.result_for_solid('solid_two').transformed_value() == 'barbar'


def test_execute_dep_solid_different_input_name():
    pass_to_first = define_pass_value_solid('pass_to_first')

    first_solid = single_output_transform(
        'first_solid',
        inputs=[InputDefinition(name='a_thing')],
        transform_fn=lambda context, inputs: inputs['a_thing'] + inputs['a_thing'],
        output=dagster.OutputDefinition(),
    )

    second_solid = single_output_transform(
        'second_solid',
        inputs=[
            InputDefinition(name='an_input'),
        ],
        transform_fn=lambda context, inputs: inputs['an_input'] + inputs['an_input'],
        output=dagster.OutputDefinition(),
    )

    pipeline = dagster.PipelineDefinition(
        solids=[pass_to_first, first_solid, second_solid],
        dependencies={
            'first_solid': {
                'a_thing': DependencyDefinition('pass_to_first'),
            },
            'second_solid': {
                'an_input': DependencyDefinition('first_solid'),
            },
        }
    )

    result = dagster.execute_pipeline(
        pipeline,
        environment=config.Environment(solids={'pass_to_first': config.Solid({
            'value': 'bar'
        })})
    )

    assert result.success
    assert len(result.result_list) == 3
    assert result.result_list[0].transformed_value() == 'bar'
    assert result.result_list[1].transformed_value() == 'barbar'
    assert result.result_list[2].transformed_value() == 'barbarbarbar'



import pytest

from dagster import (
    ConfigDefinition,
    DagsterInvariantViolationError,
    Field,
    PipelineDefinition,
    SolidDefinition,
    config,
    execute_pipeline,
    types,
)

from dagster.core.errors import DagsterTypeError


def test_basic_solid_with_config():
    did_get = {}

    def _t_fn(info, _inputs):
        did_get['yep'] = info.config

    solid = SolidDefinition(
        name='solid_with_context',
        inputs=[],
        outputs=[],
        config_def=ConfigDefinition.config_dict('SomeConfig', {'some_config': Field(types.String)}),
        transform_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solids=[solid])

    execute_pipeline(
        pipeline,
        config.Environment(solids={'solid_with_context': config.Solid({
            'some_config': 'foo'
        })}),
    )

    assert 'yep' in did_get
    assert 'some_config' in did_get['yep']


def test_config_arg_mismatch():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid = SolidDefinition(
        name='solid_with_context',
        inputs=[],
        outputs=[],
        config_def=ConfigDefinition.config_dict('SomeConfig', {'some_config': Field(types.String)}),
        transform_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solids=[solid])

    with pytest.raises(DagsterTypeError):
        execute_pipeline(
            pipeline,
            config.Environment(solids={'solid_with_context': config.Solid({
                'some_config': 1
            })}),
        )


def test_solid_not_found():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid = SolidDefinition(
        name='find_me_solid',
        inputs=[],
        outputs=[],
        transform_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solids=[solid])

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(
            pipeline,
            config.Environment(solids={
                'not_found': config.Solid({
                    'some_config': 1,
                }),
            }),
        )


def test_config_for_no_config():
    def _t_fn(*_args):
        raise Exception('should not reach')

    solid_def = SolidDefinition(
        name='no_config_solid',
        inputs=[],
        outputs=[],
        transform_fn=_t_fn,
    )

    pipeline = PipelineDefinition(solids=[solid_def])

    with pytest.raises(
        DagsterInvariantViolationError,
        match="Solid no_config_solid was provided {'some_config': 1} but does not take config",
    ):
        execute_pipeline(
            pipeline,
            config.Environment(solids={
                'no_config_solid': config.Solid({
                    'some_config': 1,
                }),
            }),
        )



from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    Field,
    PipelineContextDefinition,
    PipelineDefinition,
    types,
    InputDefinition,
    OutputDefinition,
    lambda_solid,
    solid,
)


def test_deps_equal():
    assert DependencyDefinition('foo') == DependencyDefinition('foo')
    assert DependencyDefinition('foo') != DependencyDefinition('bar')

    assert DependencyDefinition('foo', 'bar') == DependencyDefinition('foo', 'bar')
    assert DependencyDefinition('foo', 'bar') != DependencyDefinition('foo', 'quuz')


def test_pipeline_types():
    ContextOneConfigDict = types.ConfigDictionary(
        'ContextOneConfigDict',
        {'field_one': Field(types.String)},
    )

    SolidOneConfigDict = types.ConfigDictionary(
        'SolidOneConfigDict',
        {'another_field': Field(types.Int)},
    )

    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        inputs=[InputDefinition('input_one', types.String)],
        outputs=[OutputDefinition(types.Any)],
        config_def=ConfigDefinition(SolidOneConfigDict),
    )
    def solid_one(_info, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solids=[produce_string, solid_one],
        dependencies={'solid_one': {
            'input_one': DependencyDefinition('produce_string'),
        }},
        context_definitions={
            'context_one':
            PipelineContextDefinition(
                context_fn=lambda: None,
                config_def=ConfigDefinition(ContextOneConfigDict),
            )
        }
    )

    present_types = [
        SolidOneConfigDict,
        ContextOneConfigDict,
        types.String,
        types.Any,
        types.Int,
    ]

    for present_type in present_types:
        name = present_type.name
        assert pipeline_def.has_type(name)
        assert pipeline_def.type_named(name).name == name

    not_present_types = [
        types.Bool,
        types.Dict,
        types.PythonObjectType('Duisjdfke', dict),
    ]

    for not_present_type in not_present_types:
        assert not pipeline_def.has_type(not_present_type.name)



from collections import (
    defaultdict,
    namedtuple,
)
import keyword
import re
from toposort import toposort_flatten

from dagster import check
from dagster.core import types
from dagster.utils.logging import (
    level_from_string,
    define_colored_console_logger,
)

from dagster.config import DEFAULT_CONTEXT_NAME

from .errors import DagsterInvalidDefinitionError

from .execution_context import ExecutionContext

from .types import (
    ConfigDictionary,
    DagsterType,
    Field,
)

DEFAULT_OUTPUT = 'result'

DISALLOWED_NAMES = set(
    [
        'context',
        'conf',
        'meta',
        'arg_dict',
        'dict',
        'input_arg_dict',
        'output_arg_dict',
        'int',
        'str',
        'float',
        'bool',
        'input',
        'output',
        'result',
        'type',
    ] + keyword.kwlist  # just disallow all python keywords
)


def check_valid_name(name):
    check.str_param(name, 'name')
    if name in DISALLOWED_NAMES:
        raise DagsterInvalidDefinitionError('{name} is not allowed'.format(name=name))

    regex = r'^[A-Za-z0-9_]+$'
    if not re.match(regex, name):
        raise DagsterInvalidDefinitionError(
            '{name} must be in regex {regex}'.format(name=name, regex=regex)
        )
    return name


class PipelineContextDefinition(object):
    '''Pipelines declare the different context types they support, in the form
    of PipelineContextDefinitions. For example a pipeline could declare a context
    definition for different operating environments: unittest, integration tests,
    production and so forth. The use provides context function that returns an
    ExecutionContext that is passed to every solid. One can hang resources
    (such as db connections) off of that context. Thus the pipeline author
    has complete control over how the author of each individual solid within
    the pipeline interacts with its operating environment.

    The PipelineContextDefinition is passed to the PipelineDefinition in
    a dictionary key'ed by its name so the name is not present in this object.

    Attributes:
        config_def (ConfigurationDefinition): The configuration for the pipeline context.

        context_fn (callable):
            Signature is (pipeline: PipelineDefintion, config_value: Any) => ExecutionContext

            A callable that either returns *or* yields an ExecutionContext.

        description (str): A description of what this context represents
    '''

    @staticmethod
    def passthrough_context_definition(context):
        '''Create a context definition from a pre-existing context. This can be useful
        in testing contexts where you may want to create a context manually and then
        pass it into a one-off PipelineDefinition

        Args:
            context (ExecutionContext): The context that will provided to the pipeline.
        Returns:
            PipelineContextDefinition: The passthrough context definition.
        '''

        check.inst_param(context, 'context', ExecutionContext)
        context_definition = PipelineContextDefinition(context_fn=lambda *_args: context)
        return {DEFAULT_CONTEXT_NAME: context_definition}

    def __init__(self, context_fn, config_def=None, description=None):
        '''
        Args:
            context_fn (callable):
                Signature of context_fn:
                (pipeline: PipelineDefintion, config_value: Any) => ExecutionContext

                Returns *or* yields an ExecutionContext.

                If it yields a context, the code after the yield executes after pipeline
                completion, just like a python context manager.

                Environment-specific resources should be placed in the "resources" argument
                to an execution context. This argument can be *anything* and it is made
                avaiable to every solid in the pipeline. A typical pattern is to have this
                resources object be a namedtuple, where each property is an object that
                manages a particular resource, e.g. aws, a local filesystem manager, etc.

            config_def (ConfigDefinition):
                Define the configuration for the context

            description (str): Description of the context definition.
        '''
        self.config_def = check.opt_inst_param(
            config_def,
            'config_def',
            ConfigDefinition,
            ConfigDefinition(),
        )
        self.context_fn = check.callable_param(context_fn, 'context_fn')
        self.description = description


DefaultContextConfigDict = ConfigDictionary(
    'DefaultContextConfigDict',
    {
        'log_level': Field(
            dagster_type=types.String,
            is_optional=True,
            default_value='INFO',
        ),
    },
)


def _default_pipeline_context_definitions():
    def _default_context_fn(info):
        log_level = level_from_string(info.config['log_level'])
        context = ExecutionContext(
            loggers=[define_colored_console_logger('dagster', level=log_level)]
        )
        return context

    default_context_def = PipelineContextDefinition(
        config_def=ConfigDefinition(DefaultContextConfigDict),
        context_fn=_default_context_fn,
    )
    return {DEFAULT_CONTEXT_NAME: default_context_def}


class DependencyDefinition(namedtuple('_DependencyDefinition', 'solid output description')):
    '''Dependency definitions represent an edge in the DAG of solids. This object is
    used with a dictionary structure (whose keys represent solid/input where the dependency
    comes from) so this object only contains the target dependency information.

    Attributes:
        solid (str):
            The name of the solid that is the target of the dependency.
            This is the solid where the value passed between the solids
            comes from.
        output (str):
            The name of the output that is the target of the dependency.
            Defaults to "result", the default output name of solids with a single output.
        description (str):
            Description of this dependency. Optional.
    '''

    def __new__(cls, solid, output=DEFAULT_OUTPUT, description=None):
        return super(DependencyDefinition, cls).__new__(
            cls,
            check.str_param(solid, 'solid'),
            check.str_param(output, 'output'),
            check.opt_str_param(description, 'description'),
        )


class InputToOutputHandleDict(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidInputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidInputHandle)
        check.inst_param(val, 'val', SolidOutputHandle)
        return dict.__setitem__(self, key, val)


def check_two_dim_dict(ddict, param_name, key_type=None, value_type=None):
    check.dict_param(ddict, param_name, key_type=key_type, value_type=dict)
    for sub_dict in ddict.values():
        check.dict_param(sub_dict, 'sub_dict', key_type=key_type, value_type=value_type)
    return ddict


def check_opt_two_dim_dict(ddict, param_name, key_type=None, value_type=None):
    ddict = check.opt_dict_param(ddict, param_name, key_type=key_type, value_type=dict)
    for sub_dict in ddict.values():
        check.dict_param(sub_dict, 'sub_dict', key_type=key_type, value_type=value_type)
    return ddict


def check_two_dim_str_dict(ddict, param_name, value_type):
    return check_two_dim_dict(ddict, param_name, key_type=str, value_type=value_type)


def check_opt_two_dim_str_dict(ddict, param_name, value_type):
    return check_opt_two_dim_dict(ddict, param_name, key_type=str, value_type=value_type)


def _create_handle_dict(solid_dict, dep_dict):
    check.dict_param(solid_dict, 'solid_dict', key_type=str, value_type=Solid)
    check_two_dim_str_dict(dep_dict, 'dep_dict', DependencyDefinition)

    handle_dict = InputToOutputHandleDict()

    for solid_name, input_dict in dep_dict.items():
        for input_name, dep_def in input_dict.items():
            from_solid = solid_dict[solid_name]
            to_solid = solid_dict[dep_def.solid]
            handle_dict[from_solid.input_handle(input_name)] = to_solid.output_handle(
                dep_def.output
            )

    return handle_dict


class DependencyStructure(object):
    @staticmethod
    def from_definitions(solids, dep_dict):
        return DependencyStructure(_create_handle_dict(solids, dep_dict))

    def __init__(self, handle_dict):
        self._handle_dict = check.inst_param(handle_dict, 'handle_dict', InputToOutputHandleDict)

    def has_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return solid_input_handle in self._handle_dict

    def deps_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        return list(handles[1] for handles in self.__gen_deps_of_solid(solid_name))

    def deps_of_solid_with_input(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        return dict(self.__gen_deps_of_solid(solid_name))

    def __gen_deps_of_solid(self, solid_name):
        for input_handle, output_handle in self._handle_dict.items():
            if input_handle.solid.name == solid_name:
                yield (input_handle, output_handle)

    def depended_by_of_solid(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        result = defaultdict(list)
        for input_handle, output_handle in self._handle_dict.items():
            if output_handle.solid.name == solid_name:
                result[output_handle].append(input_handle)
        return result

    def get_dep(self, solid_input_handle):
        check.inst_param(solid_input_handle, 'solid_input_handle', SolidInputHandle)
        return self._handle_dict[solid_input_handle]

    def input_handles(self):
        return list(self._handle_dict.keys())

    def items(self):
        return self._handle_dict.items()


class SolidInstance(namedtuple('Solid', 'name alias')):
    '''
    A solid identifier in a dependency structure. Allows supplying parameters to the solid,
    like the alias.

    Example:

        .. code-block:: python

            pipeline = Pipeline(
                solids=[solid_1, solid_2]
                dependencies={
                    SolidInstance('solid_2', alias='other_name') : {
                        'input_name' : DependencyDefinition('solid_2'),
                    },
                    'solid_1' : {
                        'input_name': DependencyDefinition('other_name'),
                    },
                }
            )
    '''

    def __new__(cls, name, alias=None):
        name = check.str_param(name, 'name')
        alias = check.opt_str_param(alias, 'alias')
        return super(cls, SolidInstance).__new__(cls, name, alias)


class Solid(object):
    '''
    Solid instance within a pipeline. Defined by it's name inside the pipeline.

    Attributes:
        name (str):
            Name of the solid inside the pipeline. Must be unique per-pipeline.
        definition (SolidDefinition):
            Definition of the solid.
    '''

    def __init__(self, name, definition):
        self.name = name
        self.definition = definition

        input_handles = {}
        for input_def in self.definition.input_defs:
            input_handles[input_def.name] = SolidInputHandle(self, input_def)

        self._input_handles = input_handles

        output_handles = {}
        for output_def in self.definition.output_defs:
            output_handles[output_def.name] = SolidOutputHandle(self, output_def)

        self._output_handles = output_handles

    def input_handles(self):
        return self._input_handles.values()

    def output_handles(self):
        return self._output_handles.values()

    def input_handle(self, name):
        check.str_param(name, 'name')
        return self._input_handles[name]

    def output_handle(self, name):
        check.str_param(name, 'name')
        return self._output_handles[name]

    def has_input(self, name):
        return self.definition.has_input(name)

    def input_def_named(self, name):
        return self.definition.input_def_named(name)

    def has_output(self, name):
        return self.definition.has_output(name)

    def output_def_named(self, name):
        return self.definition.output_def_named(name)

    @property
    def input_defs(self):
        return self.definition.input_defs

    @property
    def output_defs(self):
        return self.definition.output_defs

    @property
    def config_def(self):
        return self.definition.config_def


class SolidAliasMapper:
    def __init__(self, dependencies_dict):
        aliased_dependencies_dict = {}
        solid_uses = defaultdict(set)

        for solid_key, input_dep_dict in dependencies_dict.items():
            if not isinstance(solid_key, SolidInstance):
                solid_key = SolidInstance(solid_key)

            if solid_key.alias:
                key = solid_key.name
                alias = solid_key.alias
            else:
                key = solid_key.name
                alias = solid_key.name

            solid_uses[key].add(alias)
            aliased_dependencies_dict[alias] = input_dep_dict

            for dependency in input_dep_dict.values():
                solid_uses[dependency.solid].add(dependency.solid)

        self.solid_uses = solid_uses
        self.aliased_dependencies_dict = aliased_dependencies_dict

    def get_uses_of_solid(self, solid_def_name):
        # For the case when solids are passed, but no dependency structure.
        if not self.aliased_dependencies_dict:
            return set([solid_def_name])

        return self.solid_uses.get(solid_def_name)


def _create_execution_structure(name, solids, dependencies_dict):
    mapper = SolidAliasMapper(dependencies_dict)

    pipeline_solids = []
    for solid_def in solids:
        if isinstance(solid_def, SolidDefinition):
            uses_of_solid = mapper.get_uses_of_solid(solid_def.name)

            if uses_of_solid is None:
                raise DagsterInvalidDefinitionError(
                    'Solid {name} is passed to list of pipeline solids, but is not used'.format(
                        name=solid_def.name
                    )
                )

            check.inst(uses_of_solid, set, 'must be a set')

            for alias in uses_of_solid:
                pipeline_solids.append(Solid(name=alias, definition=solid_def))

        elif callable(solid_def):
            raise DagsterInvalidDefinitionError(
                '''You have passed a lambda or function {func} into a pipeline that is
                not a solid. You have likely forgetten to annotate this function with
                an @solid or @lambda_solid decorator located in dagster.core.decorators
                '''.format(func=solid_def.__name__)
            )
        else:
            raise DagsterInvalidDefinitionError(
                'Invalid item in solid list: {item}'.format(item=repr(solid_def))
            )

    pipeline_solid_dict = {ps.name: ps for ps in pipeline_solids}

    _validate_dependences(mapper.aliased_dependencies_dict, pipeline_solid_dict)

    dependency_structure = DependencyStructure.from_definitions(
        pipeline_solid_dict,
        mapper.aliased_dependencies_dict,
    )

    _validate_dependency_structure(name, pipeline_solid_dict, dependency_structure)

    return dependency_structure, pipeline_solid_dict


def _validate_dependences(dependencies, solid_dict):
    for from_solid, dep_by_input in dependencies.items():
        for from_input, dep in dep_by_input.items():
            if from_solid == dep.solid:
                raise DagsterInvalidDefinitionError(
                    'Circular reference detected in solid {from_solid} input {from_input}.'.format(
                        from_solid=from_solid, from_input=from_input
                    )
                )

            if not from_solid in solid_dict:
                raise DagsterInvalidDefinitionError(
                    'Solid {from_solid} in dependency dictionary not found in solid list'.format(
                        from_solid=from_solid
                    ),
                )

            if not solid_dict[from_solid].definition.has_input(from_input):
                input_list = [
                    input_def.name for input_def in solid_dict[from_solid].definition.input_defs
                ]
                raise DagsterInvalidDefinitionError(
                    'Solid {from_solid} does not have input {from_input}. '.format(
                        from_solid=from_solid,
                        from_input=from_input,
                    ) + \
                    'Input list: {input_list}'.format(input_list=input_list)
                )

            if not dep.solid in solid_dict:
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} in DependencyDefinition not found in solid list'.format(
                        dep=dep
                    ),
                )

            if not solid_dict[dep.solid].definition.has_output(dep.output):
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} does not have output {dep.output}'.format(dep=dep),
                )


def _validate_dependency_structure(name, pipeline_solid_dict, dependency_structure):
    for pipeline_solid in pipeline_solid_dict.values():
        solid = pipeline_solid.definition
        for input_def in solid.input_defs:
            if not dependency_structure.has_dep(pipeline_solid.input_handle(input_def.name)):
                error_msg = (
                    'Dependency must be specified for solid ' +
                    '{pipeline_name} input {input_name}'.format(
                        pipeline_name=pipeline_solid.name,
                        input_name=input_def.name,
                    )
                )
                if name:
                    error_msg += ' in pipeline {name}'.format(name=name)
                raise DagsterInvalidDefinitionError(error_msg)


def _gather_all_types(solids, context_definitions):
    check.list_param(solids, 'solids', SolidDefinition)
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )

    for solid in solids:
        for dagster_type in solid.iterate_types():
            yield dagster_type

    for context_definition in context_definitions.values():
        if context_definition.config_def:
            for dagster_type in context_definition.config_def.config_type.iterate_types():
                yield dagster_type


def construct_type_dictionary(solids, context_definitions):
    type_dict = {}
    all_types = list(_gather_all_types(solids, context_definitions))
    for dagster_type in all_types:
        name = dagster_type.name
        if name in type_dict:
            if dagster_type is not type_dict[name]:
                raise DagsterInvalidDefinitionError(
                    (
                        'Type names must be unique. You have construct two instances of types '
                        'with the same name {name} but have different instances'.format(name=name)
                    )
                )
        else:
            type_dict[dagster_type.name] = dagster_type

    return type_dict


class PipelineDefinition(object):
    '''A instance of a PipelineDefinition represents a pipeline in dagster.

    A pipeline is comprised of:

    - Solids:
        Each solid represents a functional unit of data computation.

    - Context Definitions:
        Pipelines can be designed to execute in a number of different operating environments
        (e.g. prod, dev, unittest) that require different configuration and setup. A context
        definition defines how a context (of type ExecutionContext) is created and what
        configuration is necessary to create it.

    - Dependencies:
        Solids within a pipeline are arranged as a DAG (directed, acyclic graph). Dependencies
        determine how the values produced by solids flow through the DAG.

    Attributes:
        name (str):
            Name of the pipeline. Must be unique per-repository.
        description (str):
            Description of the pipeline. Optional.
        solids (List[SolidDefinition]):
            List of the solids in this pipeline.
        dependencies (Dict[str, Dict[str, DependencyDefinition]]) :
            Dependencies that constitute the structure of the pipeline. This is a two-dimensional
            array that maps solid_name => input_name => DependencyDefiniion instance
        context_definitions (Dict[str, PipelineContextDefinition]):
            The context definitions available for consumers of this pipelines. For example, a
            unit-testing environment and a production environment probably have very different
            configuration and requirements. There would be one context definition per
            environment.
        dependency_structure (DependencyStructure):
            Used mostly internally. This has the same information as the dependencies data
            structure, but indexed for fast usage.
    '''

    def __init__(
        self,
        solids,
        name=None,
        description=None,
        context_definitions=None,
        dependencies=None,
    ):
        '''
        Args:
            solids (List[SolidDefinition]): Solids in the pipeline
            name (str): Name. This is optional, mostly for situations that require ephemeral
                pipeline definitions for fast scaffolding or testing.
            description (str): Description of the pipline.
            context_definitions (Dict[str, PipelineContextDefinition]): See class description.
            dependencies: (Dict[str, Dict[str, DependencyDefinition]]): See class description.
        '''
        self.name = check.opt_str_param(name, 'name', '<<unnamed>>')
        self.description = check.opt_str_param(description, 'description')

        if context_definitions is None:
            context_definitions = _default_pipeline_context_definitions()

        self.context_definitions = check.dict_param(
            context_definitions,
            'context_definitions',
            key_type=str,
            value_type=PipelineContextDefinition,
        )

        dependencies = check_opt_two_dim_dict(
            dependencies,
            'dependencies',
            value_type=DependencyDefinition,
        )

        dependency_structure, pipeline_solid_dict = _create_execution_structure(
            name,
            solids,
            dependencies,
        )

        self._solid_dict = pipeline_solid_dict
        self.dependency_structure = dependency_structure
        self._type_dict = construct_type_dictionary(solids, self.context_definitions)

    @staticmethod
    def create_single_solid_pipeline(pipeline, solid_name, injected_solids=None):
        '''
        Return a new pipeline which is a single solid of the passed-in pipeline.

        Frequently (especially in test contexts) one wants to isolate a single solid for
        independent testing.

        See PipelineDefinition.create_sub_pipeline.

        Args:
            pipeline (PipelineDefinition): PipelineDefinition that we will subset
            solid_name (str): Name of the solid to isolate
            injected_solids (Dict[str, Dict[str, SolidDefinition]]):
                When you create a subpipeline, you possible left with solids within that pipeline
                who have unmet dependencies. To fulfill these dependencies new solids must be
                provided.

        Returns:
            PipelineDefinition: The new pipeline with only the passed-in solid and the injected
            solids.

        Example:

        .. code-block:: python

            new_pipeline = PipelineDefinition.create_single_solid_pipeline(
                existing_pipeline,
                'A', # name of solid within existing_pipeline
                {
                    'A': {
                        # new_solid_instance is a solid that outputs something
                        # that is compatible with A_input
                        'A_input': new_solid_instance,
                    },
                },
            )

        '''
        return PipelineDefinition.create_sub_pipeline(
            pipeline,
            [solid_name],
            [solid_name],
            injected_solids,
        )

    @staticmethod
    def create_sub_pipeline(pipeline, from_solids, through_solids, injected_solids=None):
        '''
        Return a new pipeline which is a subset of the passed-in pipeline.

        In addition to making sub-pipelines out of a single solid, one can also create a
        pipeline using an arbitrary subset of the dag. In this case, we start "from" a set
        of solids and then proceed forward through the dependency graph until all the
        "through" solids are reached.

        See PipelineDefinition.create_single_solid_pipeline

        Args:
            pipeline (PipelineDefinition): PipelineDefinition that we will subset
            from_solids (List[str]):
                List solids to "start" from. Inputs into these solids will have to satisfied
                via the injected solids parameter.
            through_solids (List[str]):
                List of solids to execute "through". Solids depending on these solids
                transitively will not be included in the returned PipelineDefinition.
            injected_solids (Dict[str, Dict[str, SolidDefinition]]):
                When you create a subpipeline, you possible left with solids within that pipeline
                who have unmet dependencies. To fulfill these dependencies new solids must be
                provided.

        Returns:
            PipelineDefinition:
                The new pipeline definition that contains all the solids from the "from_solids"
                through the "through_solids", plus the injected solids.
        '''
        check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        check.list_param(from_solids, 'from_solids', of_type=str)
        check.list_param(through_solids, 'through_solids', of_type=str)
        injected_solids = check_opt_two_dim_str_dict(
            injected_solids, 'injected_solids', SolidDefinition
        )

        # FIXME: fix circular reference
        subgraph = ExecutionGraph.from_pipeline_subset(
            pipeline,
            from_solids,
            through_solids,
            injected_solids,
        )

        return subgraph.to_pipeline()

    @property
    def display_name(self):
        '''Name suitable for exception messages, logging etc. If pipeline
        is unnamed the method with return "<<unnamed>>".

        Returns:
            str: Display name of pipeline
        '''
        return self.name if self.name else '<<unnamed>>'

    @property
    def solids(self):
        '''Return the solids in the pipeline.

        Returns:
            List[SolidDefinition]: List of solids.
        '''
        return list(set(self._solid_dict.values()))

    def has_solid(self, name):
        '''Return whether or not the solid is in the piepline

        Args:
            name (str): Name of solid

        Returns:
            bool: True if the solid is in the pipeline
        '''
        check.str_param(name, 'name')
        return name in self._solid_dict

    def solid_named(self, name):
        '''Return the solid named "name". Throws if it does not exist.

        Args:
            name (str): Name of solid

        Returns:
            SolidDefinition: SolidDefinition with correct name.
        '''
        check.str_param(name, 'name')
        return self._solid_dict[name]

    def has_type(self, name):
        check.str_param(name, 'name')
        return name in self._type_dict

    def type_named(self, name):
        check.str_param(name, 'name')
        return self._type_dict[name]


class ExpectationResult(object):
    '''
    When Expectations are evaluated in the callback passed to ExpectationDefinitions,
    the user must return an ExpectationResult object from the callback.

    Attributes:

        success (bool): Whether the expectation passed or not.
        message (str): Information about the computation. Typically only used in the failure case.
        result_context (Any): Arbitrary information about the expectation result.
    '''

    def __init__(self, success, message=None, result_context=None):
        self.success = check.bool_param(success, 'success')
        self.message = check.opt_str_param(message, 'message')
        self.result_context = check.opt_dict_param(result_context, 'result_context')


class ExpectationDefinition(object):
    '''
    Expectations represent a data quality test. It performs an arbitrary computation
    to see if a given input or output satisfies the expectation.

    Attributes:

        name (str): The name of the expectation. Names should be unique per-solid.
        expectation_fn (callable):
            This is the implementation of expectation computation. It should be a callback
            of the form.

            (context: ExecutionContext, info: ExpectationExecutionInfo, value: Any)
            : ExpectationResult

            "value" conforms to the type check performed within the Dagster type system.

            e.g. If the expectation is declare on an input of type dagster_pd.DataFrame, you can
            assume that value is a pandas.DataFrame

        description (str): Description of expectation. Optional.

    Example:

    .. code-block:: python

        InputDefinition('some_input', types.Int, expectations=[
            ExpectationDefinition(
                name='is_positive',
                expectation_fn=lambda(
                    _info,
                    value,
                ): ExpectationResult(success=value > 0),
            )
        ])
    '''

    def __init__(self, name, expectation_fn, description=None):
        self.name = check_valid_name(name)
        self.expectation_fn = check.callable_param(expectation_fn, 'expectation_fn')
        self.description = check.opt_str_param(description, 'description')


class InputDefinition(object):
    '''An InputDefinition instance represents an argument to a transform defined within a solid.
    Inputs are values within the dagster type system that are created from previous solids.

    Attributes:
        name (str): Name of the input.
        dagster_type (DagsterType): Type of the input. Defaults to types.Any
        expectations (List[ExpectationDefinition]):
            List of expectations that applies to the value passed to the solid.
        description (str): Description of the input. Optional.
    '''

    def __init__(self, name, dagster_type=types.Any, expectations=None, description=None):
        ''
        self.name = check_valid_name(name)

        self.dagster_type = check.inst_param(dagster_type, 'dagster_type', types.DagsterType)

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
        self.description = check.opt_str_param(description, 'description')

    @property
    def descriptive_key(self):
        return 'output'


class OutputDefinition(object):
    '''An OutputDefinition represents an output from a solid. Solids can have multiple
    outputs. In those cases the outputs must be named. Frequently solids have only one
    output, and so the user can construct a single OutputDefinition that will have
    the default name of "result".

    Attributes:
        dagster_type (DagsterType): Type of the output. Defaults to types.Any.
        name (str): Name of the output. Defaults to "result".
        expectations List[ExpectationDefinition]: Expectations for this output.
        description (str): Description of the output. Optional.
    '''

    def __init__(self, dagster_type=None, name=None, expectations=None, description=None):
        self.name = check.opt_str_param(name, 'name', DEFAULT_OUTPUT)

        self.dagster_type = check.opt_inst_param(
            dagster_type, 'dagster_type', types.DagsterType, types.Any
        )

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
        self.description = check.opt_str_param(description, 'description')

    @property
    def descriptive_key(self):
        return 'output'


def _kv_str(key, value):
    return '{key}="{value}"'.format(key=key, value=repr(value))


def struct_to_string(name, **kwargs):
    props_str = ', '.join([_kv_str(key, value) for key, value in kwargs.items()])
    return '{name}({props_str})'.format(name=name, props_str=props_str)


class SolidInputHandle(namedtuple('_SolidInputHandle', 'solid input_def')):
    def __new__(cls, solid, input_def):
        return super(SolidInputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(input_def, 'input_def', InputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidInputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            input_name=self.input_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.input_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.input_def.name == other.input_def.name


class SolidOutputHandle(namedtuple('_SolidOutputHandle', 'solid output_def')):
    def __new__(cls, solid, output_def):
        return super(SolidOutputHandle, cls).__new__(
            cls,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(output_def, 'output_def', OutputDefinition),
        )

    def _inner_str(self):
        return struct_to_string(
            'SolidOutputHandle',
            solid_name=self.solid.name,
            definition_name=self.solid.definition.name,
            output_name=self.output_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.solid.name, self.output_def.name))

    def __eq__(self, other):
        return self.solid.name == other.solid.name and self.output_def.name == other.output_def.name


class Result(namedtuple('_Result', 'value output_name')):
    '''A solid transform function return a stream of Result objects.
    An implementator of a SolidDefinition must provide a transform that
    yields objects of this type.

    Attributes:
        value (Any): Value returned by the transform.
        output_name (str): Name of the output returns. defaults to "result"
'''

    def __new__(cls, value, output_name=DEFAULT_OUTPUT):
        return super(Result, cls).__new__(
            cls,
            value,
            check.str_param(output_name, 'output_name'),
        )


class ConfigDefinition(object):
    '''Represents the configuration of an entity in Dagster

    Broadly defined, configs determine how computations within dagster interact with
    the external world. Example values that would end up in configs would be file paths,
    database table names, and so forth.

    Attributes:

        config_type (DagsterType): Type of the config.
    '''

    @staticmethod
    def config_dict(name, field_dict):
        '''Shortcut to create a dictionary based config definition.


        Args:
            field_dict (dict): dictionary of `Field` objects keyed by their names.

        Example:

        .. code-block:: python

            ConfigDefinition.config_dict({
                'int_field': Field(types.Int),
                'string_field': Field(types.String),
             })

        '''
        return ConfigDefinition(types.ConfigDictionary(name, field_dict))

    def __init__(self, config_type=types.Any, description=None):
        '''Construct a ConfigDefinition

        Args:
            config_type (DagsterType): Type the determines shape and values of config'''
        self.config_type = check.inst_param(config_type, 'config_type', DagsterType)
        self.description = check.opt_str_param(description, 'description')


class SolidDefinition(object):
    '''A solid (a name extracted from the acronym of "software-structured data" (SSD)) represents
    a unit of computation within a data pipeline.

    As its core, a solid is a function. It accepts inputs (which are values produced from
    other solids) and configuration, and produces outputs. These solids are composed as a
    directed, acyclic graph (DAG) within a pipeline to form a computation that produces
    data assets.

    Solids should be implemented as idempotent, parameterizable, non-destructive functions.
    Data computations with these properties are much easier to test, reason about, and operate.

    The inputs and outputs are gradually, optionally typed by the dagster type system. Types
    can be user-defined and can represent entites as varied as scalars, dataframe, database
    tables, and so forth. They can represent pure in-memory objects, or handles to assets
    on disk or in external resources.

    A solid is a generalized abstraction that could take many forms.

    Example:

        .. code-block:: python

            def _read_csv(info, inputs):
                yield Result(pandas.read_csv(info.config['path']))

            SolidDefinition(
                name='read_csv',
                inputs=[],
                config_def=ConfigDefinition(types.ConfigDictionary({'path' => types.Path})),
                outputs=[OutputDefinition()] # default name ('result') and any typed
                transform_fn
            )

    Attributes:
        name (str): Name of the solid.
        inputs (List[InputDefiniton]): Inputs of the solid.
        transform_fn (callable):
            Callable with the signature
            (
                info: TransformExecutionInfo,
                inputs: Dict[str, Any],
            ) : Iterable<Result>
        outputs (List[OutputDefinition]): Outputs of the solid.
        config_def (ConfigDefinition): How the solid configured.
        description (str): Description of the solid.
        metadata (dict):
            Arbitrary metadata for the solid. Some frameworks expect and require
            certain metadata to be attached to a solid.
    '''

    def __init__(
        self,
        name,
        inputs,
        transform_fn,
        outputs,
        config_def=None,
        description=None,
        metadata=None,
    ):
        self.name = check_valid_name(name)
        self.input_defs = check.list_param(inputs, 'inputs', InputDefinition)
        self.transform_fn = check.callable_param(transform_fn, 'transform_fn')
        self.output_defs = check.list_param(outputs, 'outputs', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')
        self.config_def = check.opt_inst_param(
            config_def,
            'config_def',
            ConfigDefinition,
        )
        self.metadata = check.opt_dict_param(metadata, 'metadata', key_type=str)
        self._input_dict = {inp.name: inp for inp in inputs}
        self._output_dict = {output.name: output for output in outputs}

    def has_input(self, name):
        check.str_param(name, 'name')
        return name in self._input_dict

    def input_def_named(self, name):
        check.str_param(name, 'name')
        return self._input_dict[name]

    def has_output(self, name):
        check.str_param(name, 'name')
        return name in self._output_dict

    def output_def_named(self, name):
        check.str_param(name, 'name')
        return self._output_dict[name]

    def iterate_types(self):
        for input_def in self.input_defs:
            for dagster_type in input_def.dagster_type.iterate_types():
                yield dagster_type

        for output_def in self.output_defs:
            for dagster_type in output_def.dagster_type.iterate_types():
                yield dagster_type

        if self.config_def:
            for dagster_type in self.config_def.config_type.iterate_types():
                yield dagster_type


def _create_adjacency_lists(solids, dep_structure):
    check.list_param(solids, 'solids', Solid)
    check.inst_param(dep_structure, 'dep_structure', DependencyStructure)

    visit_dict = {s.name: False for s in solids}
    forward_edges = {s.name: set() for s in solids}
    backward_edges = {s.name: set() for s in solids}

    def visit(solid_name):
        if visit_dict[solid_name]:
            return

        visit_dict[solid_name] = True

        for output_handle in dep_structure.deps_of_solid(solid_name):
            forward_node = output_handle.solid.name
            backward_node = solid_name
            if forward_node in forward_edges:
                forward_edges[forward_node].add(backward_node)
                backward_edges[backward_node].add(forward_node)
                visit(forward_node)

    for s in solids:
        visit(s.name)

    return (forward_edges, backward_edges)


def _dependency_structure_to_dep_dict(dependency_structure):
    dep_dict = defaultdict(dict)
    for input_handle, output_handle in dependency_structure.items():
        solid_instance = SolidInstance(input_handle.solid.definition.name, input_handle.solid.name)
        dep_dict[solid_instance][input_handle.input_def.name] = DependencyDefinition(
            solid=output_handle.solid.name,
            output=output_handle.output_def.name,
        )
    return dep_dict


class ExecutionGraph(object):
    @staticmethod
    def from_pipeline(pipeline):
        check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        return ExecutionGraph(pipeline, pipeline.solids, pipeline.dependency_structure)

    @staticmethod
    def from_pipeline_subset(pipeline, from_solids, through_solids, injected_solids):
        check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        check.list_param(from_solids, 'from_solids', of_type=str)
        check.list_param(through_solids, 'through_solids', of_type=str)
        graph = ExecutionGraph.from_pipeline(pipeline)
        return _create_subgraph(graph, from_solids, through_solids).augment(injected_solids)

    def to_pipeline(self):
        return PipelineDefinition(
            solids=[solid.definition for solid in self.solids],
            dependencies=_dependency_structure_to_dep_dict(self.dependency_structure),
            context_definitions=self.pipeline.context_definitions
        )

    def augment(self, injected_solids):

        new_deps = defaultdict(dict)
        new_solids = []

        for from_solid_name, targets_by_input in injected_solids.items():
            for from_input_name, target_solid in targets_by_input.items():
                new_solids.append(Solid(name=target_solid.name, definition=target_solid))
                new_deps[from_solid_name][from_input_name] = DependencyDefinition(
                    solid=target_solid.name
                )

        check.list_param(new_solids, 'new_solids', of_type=Solid)

        solids = self.solids + new_solids

        solid_dict = {s.name: s for s in solids}

        handle_dict = InputToOutputHandleDict()
        for input_handle, output_handle in self.dependency_structure.items():
            handle_dict[input_handle] = output_handle

        for input_handle, output_handle in _create_handle_dict(solid_dict, new_deps).items():
            handle_dict[input_handle] = output_handle

        return ExecutionGraph(self.pipeline, solids, DependencyStructure(handle_dict))

    def __init__(self, pipeline, solids, dependency_structure):
        self.pipeline = pipeline
        solids = check.list_param(solids, 'solids', of_type=Solid)
        self.dependency_structure = check.inst_param(
            dependency_structure, 'dependency_structure', DependencyStructure
        )

        self._solid_dict = {s.name: s for s in solids}

        for input_handle in dependency_structure.input_handles():
            check.invariant(input_handle.solid.name in self._solid_dict)

        self.forward_edges, self.backward_edges = _create_adjacency_lists(
            solids, self.dependency_structure
        )
        self.topological_order = toposort_flatten(self.backward_edges, sort=True)

        self._transitive_deps = {}

    @property
    def topological_solids(self):
        return [self._solid_dict[name] for name in self.topological_order]

    @property
    def solids(self):
        return list(self._solid_dict.values())

    def solid_named(self, name):
        check.str_param(name, 'name')
        return self._solid_dict[name]

    def transitive_dependencies_of(self, solid_name):
        check.str_param(solid_name, 'solid_name')

        if solid_name in self._transitive_deps:
            return self._transitive_deps[solid_name]

        trans_deps = set()
        pipeline_solid = self._solid_dict[solid_name]
        solid = pipeline_solid.definition
        for input_def in solid.input_defs:
            input_handle = pipeline_solid.input_handle(input_def.name)
            if self.dependency_structure.has_dep(input_handle):
                output_handle = self.dependency_structure.get_dep(input_handle)
                trans_deps.add(output_handle.solid.name)
                trans_deps.union(self.transitive_dependencies_of(output_handle.solid.name))

        self._transitive_deps[solid_name] = trans_deps
        return self._transitive_deps[solid_name]

    def _check_solid_name(self, solid_name):
        check.str_param(solid_name, 'output_name')
        check.param_invariant(
            solid_name in self._solid_dict, 'output_name',
            'Solid {solid_name} must exist in {solid_names}'.format(
                solid_name=solid_name, solid_names=list(self._solid_dict.keys())
            )
        )

    def create_execution_subgraph(self, from_solids, to_solids):
        check.list_param(from_solids, 'from_solids', of_type=str)
        check.list_param(to_solids, 'to_solids', of_type=str)

        involved_solid_set = self._compute_involved_solid_set(from_solids, to_solids)

        involved_solids = [self._solid_dict[name] for name in involved_solid_set]

        handle_dict = InputToOutputHandleDict()

        for pipeline_solid in involved_solids:
            solid = pipeline_solid.definition
            for input_def in solid.input_defs:
                input_handle = pipeline_solid.input_handle(input_def.name)
                if self.dependency_structure.has_dep(input_handle):
                    handle_dict[input_handle] = self.dependency_structure.get_dep(input_handle)

        return ExecutionGraph(self.pipeline, involved_solids, DependencyStructure(handle_dict))

    def _compute_involved_solid_set(self, from_solids, to_solids):
        from_solid_set = set(from_solids)
        involved_solid_set = from_solid_set

        def visit(solid):
            if solid.name in involved_solid_set:
                return

            involved_solid_set.add(solid.name)

            for input_def in solid.definition.input_defs:
                input_handle = solid.input_handle(input_def.name)
                if not self.dependency_structure.has_dep(input_handle):
                    continue

                output_handle = self.dependency_structure.get_dep(input_handle)

                next_solid = output_handle.solid.name
                if next_solid in from_solid_set:
                    continue

                visit(self._solid_dict[next_solid])

        for to_solid in to_solids:
            visit(self._solid_dict[to_solid])

        return involved_solid_set


def _all_depended_on_solids(execution_graph):
    check.inst_param(execution_graph, 'execution_graph', ExecutionGraph)

    dependency_structure = execution_graph.dependency_structure

    for solid in execution_graph.solids:
        for input_def in solid.input_defs:
            input_handle = solid.input_handle(input_def.name)
            if dependency_structure.has_dep(input_handle):
                output_handle = dependency_structure.get_dep(input_handle)
                yield execution_graph.solid_named(output_handle.solid.name)


def _all_sink_solids(execution_graph):
    check.inst_param(execution_graph, 'execution_graph', ExecutionGraph)
    all_names = set([solid.name for solid in execution_graph.solids])
    all_depended_on_names = set([solid.name for solid in _all_depended_on_solids(execution_graph)])
    return all_names.difference(all_depended_on_names)


def _create_subgraph(execution_graph, from_solids, through_solids):
    check.inst_param(execution_graph, 'execution_graph', ExecutionGraph)
    check.opt_list_param(from_solids, 'from_solids', of_type=str)
    check.opt_list_param(through_solids, 'through_solids', of_type=str)

    if not through_solids:
        through_solids = list(_all_sink_solids(execution_graph))

    if not from_solids:
        all_deps = set()
        for through_solid in through_solids:
            all_deps.union(execution_graph.transitive_dependencies_of(through_solid))

        from_solids = list(all_deps)

    return execution_graph.create_execution_subgraph(from_solids, through_solids)


class RepositoryDefinition(object):
    '''Define a repository that contains a collection of pipelines.

    Attributes:
        name (str): The name of the pipeline.
        pipeline_dict (Dict[str, callable]):
            An dictionary of pipelines. The value of the dictionary is a function that takes
            no parameters and returns a PipelineDefiniton.

            We pass callables instead of the PipelineDefinitions itself so that they can be
            created on demand when accessed by name.

            As the pipelines are retrieved it ensures that the keys of the dictionary and the
            name of the pipeline are the same.

    '''

    def __init__(self, name, pipeline_dict):
        '''
        Args:
            name (str): Name of pipeline.
            pipeline_dict (Dict[str, callable]): See top-level class documentation
        '''
        self.name = check.str_param(name, 'name')

        check.dict_param(
            pipeline_dict,
            'pipeline_dict',
            key_type=str,
        )

        for val in pipeline_dict.values():
            check.is_callable(val, 'Value in pipeline_dict must be function')

        self.pipeline_dict = pipeline_dict

        self._pipeline_cache = {}

    def get_pipeline(self, name):
        '''Get a pipeline by name. Only constructs that pipeline and caches it.

        Args:
            name (str): Name of the pipeline to retriever

        Returns:
            PipelineDefinition: Instance of PipelineDefinition with that name.
'''
        if name in self._pipeline_cache:
            return self._pipeline_cache[name]

        pipeline = self.pipeline_dict[name]()
        check.invariant(
            pipeline.name == name,
            'Name does not match. Name in dict {name}. Name in pipeline {pipeline.name}'.format(
                name=name, pipeline=pipeline
            )
        )

        self._pipeline_cache[name] = check.inst(
            pipeline,
            PipelineDefinition,
            'Function passed into pipeline_dict with key {key} must return a PipelineDefinition'.
            format(key=name),
        )

        return pipeline

    def iterate_over_pipelines(self):
        '''Yield all pipelines one at a time

        Returns:
            Iterable[PipelineDefinition]:
        '''
        for name in self.pipeline_dict.keys():
            yield self.get_pipeline(name)

    def get_all_pipelines(self):
        '''Return all pipelines as a list

        Returns:
            List[PipelineDefinition]:

        '''
        return list(self.iterate_over_pipelines())


class ContextCreationExecutionInfo(
    namedtuple('_ContextCreationExecutionInfo', 'config pipeline_def')
):
    def __new__(cls, config, pipeline_def):
        return super(ContextCreationExecutionInfo, cls).__new__(
            cls,
            config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
        )


ExpectationExecutionInfo = namedtuple(
    'ExpectationExecutionInfo',
    'context inout_def solid_def expectation_def',
)


class TransformExecutionInfo(namedtuple('_TransformExecutionInfo', 'context config solid_def')):
    '''An instance of TransformExecutionInfo is passed every solid transform function.

    Attributes:

        context (ExecutionContext): Context instance for this pipeline invocation
        config (Any): Config object for current solid
    '''

    def __new__(cls, context, config, solid_def):
        return super(TransformExecutionInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', ExecutionContext),
            config,
            check.inst_param(solid_def, 'solid_def', SolidDefinition),
        )



from __future__ import print_function

import pytest

from dagster.cli.pipeline import (
    execute_print_command,
    execute_list_command,
    execute_execute_command,
)
from dagster.cli.dynamic_loader import InvalidRepositoryLoadingComboError

from dagster.utils import script_relative_path


def no_print(_):
    return None


def test_list_command():
    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        no_print,
    )

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': None,
            'module_name': 'dagster.cli.cli_tests.test_dynamic_loader',
            'fn_name': 'define_bar_repo',
        },
        no_print,
    )

    execute_list_command(
        {
            'repository_yaml': script_relative_path('repository.yml'),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        no_print,
    )

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': 'foo.py',
                'module_name': 'dagster.cli.cli_tests.test_dynamic_loader',
                'fn_name': 'define_bar_repo',
            },
            no_print,
        )

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': None,
                'module_name': 'dagster.cli.cli_tests.test_dynamic_loader',
                'fn_name': None,
            },
            no_print,
        )

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': script_relative_path('test_dynamic_loader.py'),
                'module_name': None,
                'fn_name': None,
            },
            no_print,
        )


def valid_cli_args():
    return [
        {
            'repository_yaml': script_relative_path('repository_file.yml'),
            'pipeline_name': ('foo', ),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': script_relative_path('repository_module.yml'),
            'pipeline_name': ('foo', ),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('foo', ),
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('foo', ),
            'python_file': None,
            'module_name': 'dagster.cli.cli_tests.test_dynamic_loader',
            'fn_name': 'define_bar_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': None,
            'module_name': 'dagster.cli.cli_tests.test_dynamic_loader',
            'fn_name': 'define_foo_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_foo_pipeline',
        },
    ]


def test_print_command():
    for cli_args in valid_cli_args():
        execute_print_command(
            verbose=True,
            cli_args=cli_args,
            print_fn=no_print,
        )

    for cli_args in valid_cli_args():
        execute_print_command(
            verbose=False,
            cli_args=cli_args,
            print_fn=no_print,
        )


def test_execute_command():
    for cli_args in valid_cli_args():
        execute_execute_command(
            env=None,
            cli_args=cli_args,
            print_fn=print,
        )

    for cli_args in valid_cli_args():
        execute_execute_command(
            env=script_relative_path('env.yml'),
            cli_args=cli_args,
            print_fn=print,
        )



from collections import namedtuple

from dagster import check
from dagster.utils import load_yaml_from_path

DEFAULT_CONTEXT_NAME = 'default'


# lifted from https://bit.ly/2HcQAuv
class Context(namedtuple('ContextData', 'name config')):
    def __new__(cls, name=None, config=None):
        return super(Context, cls).__new__(
            cls,
            check.opt_str_param(name, 'name', DEFAULT_CONTEXT_NAME),
            config,
        )


class Solid(namedtuple('Solid', 'config')):
    def __new__(cls, config):
        return super(Solid, cls).__new__(cls, config)


class Environment(namedtuple('EnvironmentData', 'context solids expectations')):
    def __new__(cls, solids=None, context=None, expectations=None):
        check.opt_inst_param(context, 'context', Context)
        check.opt_inst_param(expectations, 'expectations', Expectations)

        if context is None:
            context = Context()

        if expectations is None:
            expectations = Expectations(evaluate=True)

        return super(Environment, cls).__new__(
            cls,
            context=context,
            solids=check.opt_dict_param(solids, 'solids', key_type=str, value_type=Solid),
            expectations=expectations,
        )


class Expectations(namedtuple('ExpectationsData', 'evaluate')):
    def __new__(cls, evaluate):
        return super(Expectations, cls).__new__(
            cls,
            evaluate=check.bool_param(evaluate, 'evaluate'),
        )


def _construct_context(yml_config_object):
    context_obj = check.opt_dict_elem(yml_config_object, 'context')
    if context_obj:
        return Context(
            check.opt_str_elem(context_obj, 'name'),
            context_obj.get('config'),
        )
    else:
        return None


def _coerce_to_list(value):
    if value is None:
        return None
    elif isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        return value

    check.invariant('should not get here')


def _construct_solids(yml_config_object):
    solid_dict = check.opt_dict_elem(yml_config_object, 'solids')
    if solid_dict is None:
        return None

    solid_configs = {}
    for solid_name, solid_yml_object in solid_dict.items():
        config_value = solid_yml_object['config']
        solid_configs[solid_name] = Solid(config_value)

    return solid_configs


def construct_environment(yml_config_object):
    yml_config_object = check.opt_dict_param(yml_config_object, 'yml_config_object')

    return Environment(
        solids=_construct_solids(yml_config_object),
        context=_construct_context(yml_config_object),
    )


def load_environment(path):
    return construct_environment(load_yaml_from_path(path))



import pytest

from dagster import (
    ConfigDefinition,
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    ExecutionContext,
    InputDefinition,
    MultipleResults,
    OutputDefinition,
    PipelineDefinition,
    Result,
    config,
    execute_pipeline,
    lambda_solid,
    solid,
)

from dagster.core.test_utils import execute_single_solid
from dagster.core.utility_solids import define_stub_solid

# This file tests a lot of parameter name stuff
# So these warnings are spurious
# unused variables, unused arguments
# pylint: disable=W0612, W0613


def create_test_context():
    return ExecutionContext()


def create_empty_test_env():
    return config.Environment()


def test_multiple_single_result():
    mr = MultipleResults(Result('value', 'output_one'))
    assert mr.results == [Result('value', 'output_one')]


def test_multiple_double_result():
    mr = MultipleResults(Result('value_one', 'output_one'), Result('value_two', 'output_two'))
    assert mr.results == [Result('value_one', 'output_one'), Result('value_two', 'output_two')]


def test_multiple_dict():
    mr = MultipleResults.from_dict({'output_one': 'value_one', 'output_two': 'value_two'})
    assert set(mr.results) == set(
        [Result('value_one', 'output_one'),
         Result('value_two', 'output_two')]
    )


def test_no_parens_solid():
    called = {}

    @lambda_solid
    def hello_world():
        called['yup'] = True

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert called['yup']


def test_empty_solid():
    called = {}

    @lambda_solid()
    def hello_world():
        called['yup'] = True

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert called['yup']


def test_solid():
    @solid(outputs=[OutputDefinition()])
    def hello_world(_info):
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_one_output():
    @lambda_solid
    def hello_world():
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_yield():
    @solid(outputs=[OutputDefinition()])
    def hello_world(_info):
        yield Result(value={'foo': 'bar'})

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_result_return():
    @solid(outputs=[OutputDefinition()])
    def hello_world(_info):
        return Result(value={'foo': 'bar'})

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_multiple_outputs():
    @solid(outputs=[OutputDefinition(name="left"), OutputDefinition(name="right")])
    def hello_world(_info):
        return MultipleResults(
            Result(value={'foo': 'left'}, output_name='left'),
            Result(value={'foo': 'right'}, output_name='right')
        )

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    solid_result = result.result_list[0]
    assert solid_result.transformed_value('left')['foo'] == 'left'
    assert solid_result.transformed_value('right')['foo'] == 'right'


def test_dict_multiple_outputs():
    @solid(outputs=[OutputDefinition(name="left"), OutputDefinition(name="right")])
    def hello_world(_info):
        return MultipleResults.from_dict({
            'left': {
                'foo': 'left'
            },
            'right': {
                'foo': 'right'
            },
        })

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    solid_result = result.result_list[0]
    assert solid_result.transformed_value('left')['foo'] == 'left'
    assert solid_result.transformed_value('right')['foo'] == 'right'


def test_lambda_solid_with_name():
    @lambda_solid(name="foobar")
    def hello_world():
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_with_name():
    @solid(name="foobar", outputs=[OutputDefinition()])
    def hello_world(_info):
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_list[0].transformed_value()['foo'] == 'bar'


def test_solid_with_input():
    @lambda_solid(inputs=[InputDefinition(name="foo_to_foo")])
    def hello_world(foo_to_foo):
        return foo_to_foo

    pipeline = PipelineDefinition(
        solids=[define_stub_solid('test_value', {'foo': 'bar'}), hello_world],
        dependencies={'hello_world': {
            'foo_to_foo': DependencyDefinition('test_value'),
        }}
    )

    pipeline_result = execute_pipeline(pipeline)

    result = pipeline_result.result_for_solid('hello_world')

    assert result.success
    assert result.transformed_value()['foo'] == 'bar'


def test_lambda_solid_definition_errors():
    with pytest.raises(DagsterInvalidDefinitionError, match='positional vararg'):

        @lambda_solid(inputs=[InputDefinition(name="foo")])
        def vargs(foo, *args):
            pass


def test_solid_definition_errors():
    with pytest.raises(DagsterInvalidDefinitionError, match='positional vararg'):

        @solid(inputs=[InputDefinition(name="foo")], outputs=[OutputDefinition()])
        def vargs(info, foo, *args):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(inputs=[InputDefinition(name="foo")], outputs=[OutputDefinition()])
        def wrong_name(info, bar):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[
                InputDefinition(name="foo"),
                InputDefinition(name="bar"),
            ],
            outputs=[OutputDefinition()]
        )
        def wrong_name_2(info, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(inputs=[InputDefinition(name="foo")], outputs=[OutputDefinition()])
        def no_info(foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(inputs=[InputDefinition(name="foo")], outputs=[OutputDefinition()])
        def extras(_info, foo, bar):
            pass

    @solid(
        inputs=[InputDefinition(name="foo"),
                InputDefinition(name="bar")],
        outputs=[OutputDefinition()]
    )
    def valid_kwargs(info, **kwargs):
        pass

    @solid(
        inputs=[InputDefinition(name="foo"),
                InputDefinition(name="bar")],
        outputs=[OutputDefinition()]
    )
    def valid(info, foo, bar):
        pass


def test_wrong_argument_to_pipeline():
    def non_solid_func():
        pass

    with pytest.raises(
        DagsterInvalidDefinitionError, match='You have passed a lambda or function non_solid_func'
    ):
        PipelineDefinition(solids=[non_solid_func])

    with pytest.raises(
        DagsterInvalidDefinitionError, match='You have passed a lambda or function <lambda>'
    ):
        PipelineDefinition(solids=[lambda x: x])


def test_descriptions():
    @solid(description='foo')
    def solid_desc():
        pass

    assert solid_desc.description == 'foo'


def test_any_config_definition():
    called = {}
    conf_value = 234

    @solid(config_def=ConfigDefinition())
    def hello_world(info):
        assert info.config == conf_value
        called['yup'] = True

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(solids={'hello_world': config.Solid(conf_value)})
    )

    assert called['yup']



from dagster import (
    DependencyDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    config,
    execute_pipeline,
)

from dagster.sqlalchemy.templated import (
    _render_template_string,
    create_templated_sql_transform_solid,
)

from dagster.core.utility_solids import define_stub_solid

from .math_test_db import in_mem_context


def _load_table(context, table_name):
    return context.resources.sa.engine.connect().execute(
        'SELECT * FROM {table_name}'.format(table_name=table_name)
    ).fetchall()


def pipeline_test_def(solids, context, dependencies=None):
    return PipelineDefinition(
        solids=solids,
        context_definitions={
            'default': PipelineContextDefinition(context_fn=lambda info: context),
        },
        dependencies=dependencies,
    )


def test_single_templated_sql_solid_single_table_with_api():

    sql = '''CREATE TABLE {{sum_table}} AS
    SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_table_arg = 'specific_sum_table'

    sum_table_transform = create_templated_sql_transform_solid(
        name='sum_table_transform',
        sql=sql,
        table_arguments=['sum_table'],
    )

    pipeline = pipeline_test_def(solids=[sum_table_transform], context=in_mem_context())

    environment = config.Environment(
        solids={'sum_table_transform': config.Solid({
            'sum_table': sum_table_arg
        })}
    )

    result = execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_single_templated_sql_solid_double_table_raw_api():
    sum_table_arg = 'specific_sum_table'
    num_table_arg = 'specific_num_table'

    sql = '''CREATE TABLE {{sum_table}} AS
        SELECT num1, num2, num1 + num2 as sum FROM {{num_table}}'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sql,
        table_arguments=['sum_table', 'num_table'],
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context(num_table_arg))

    environment = config.Environment(
        solids={
            'sum_solid': config.Solid({
                'sum_table': sum_table_arg,
                'num_table': num_table_arg,
            })
        }
    )

    result = execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_single_templated_sql_solid_double_table_with_api():
    sum_table_arg = 'specific_sum_table'
    num_table_arg = 'specific_num_table'

    sql = '''CREATE TABLE {{sum_table}} AS
    SELECT num1, num2, num1 + num2 as sum FROM {{num_table}}'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sql,
        table_arguments=['sum_table', 'num_table'],
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context(num_table_arg))

    environment = config.Environment(
        solids={
            'sum_solid': config.Solid({
                'sum_table': sum_table_arg,
                'num_table': num_table_arg,
            })
        }
    )

    result = execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_templated_sql_solid_pipeline():
    sum_sql_template = '''CREATE TABLE {{sum_table}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_sq_sql_template = '''CREATE TABLE {{sum_sq_table}} AS
        SELECT num1, num2, sum, sum * sum as sum_sq FROM {{sum_table}}'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_table',
        sql=sum_sql_template,
        table_arguments=['sum_table'],
    )

    sum_sq_solid = create_templated_sql_transform_solid(
        name='sum_sq_table',
        sql=sum_sq_sql_template,
        table_arguments=['sum_table', 'sum_sq_table'],
        dependant_solids=[sum_solid],
    )

    context = in_mem_context()

    pipeline = pipeline_test_def(
        solids=[sum_solid, sum_sq_solid],
        context=context,
        dependencies={
            sum_sq_solid.name: {
                sum_solid.name: DependencyDefinition(sum_solid.name),
            },
        },
    )
    first_sum_table = 'first_sum_table'
    first_sum_sq_table = 'first_sum_sq_table'

    environment_one = config.Environment(
        solids={
            'sum_table':
            config.Solid({
                'sum_table': first_sum_table
            }),
            'sum_sq_table':
            config.Solid({
                'sum_table': first_sum_table,
                'sum_sq_table': first_sum_sq_table,
            }),
            #  {
            #     'sum_table': table_name_source(first_sum_table)
            # },
            # 'sum_sq_table': {
            #     'sum_sq_table': table_name_source(first_sum_sq_table)
            # },
        }
    )
    first_result = execute_pipeline(pipeline, environment=environment_one)
    assert first_result.success

    assert len(first_result.result_list) == 2
    assert first_result.result_list[0].transformed_value() == {'sum_table': first_sum_table}
    assert first_result.result_list[1].transformed_value() == {
        'sum_table': first_sum_table,
        'sum_sq_table': first_sum_sq_table,
    }

    assert _load_table(first_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]

    assert _load_table(first_result.context, first_sum_sq_table) == [(1, 2, 3, 9), (3, 4, 7, 49)]

    # now execute subdag

    pipeline_two = pipeline_test_def(
        solids=[define_stub_solid('pass_value', 'TODO'), sum_sq_solid],
        context=context,
        dependencies={
            sum_sq_solid.name: {
                sum_solid.name: DependencyDefinition('pass_value'),
            },
        },
    )

    second_sum_sq_table = 'second_sum_sq_table'

    sum_sq_args = {
        'sum_table': first_sum_table,
        'sum_sq_table': second_sum_sq_table,
    }
    environment_two = config.Environment(
        solids={
            'sum_sq_table': config.Solid(sum_sq_args),
        },
    )

    second_result = execute_pipeline(
        pipeline_two,
        environment=environment_two,
    )
    assert second_result.success
    assert len(second_result.result_list) == 2
    assert _load_table(second_result.context, second_sum_sq_table) == [(1, 2, 3, 9), (3, 4, 7, 49)]


def test_templated_sql_solid_with_api():
    sql_template = '''CREATE TABLE {{sum_table}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_solid',
        sql=sql_template,
        table_arguments=['sum_table'],
    )

    pipeline = pipeline_test_def(solids=[sum_solid], context=in_mem_context())

    sum_table_arg = 'specific_sum_table'
    environment = config.Environment(
        solids={'sum_solid': config.Solid({
            'sum_table': sum_table_arg
        })},
    )
    result = execute_pipeline(pipeline, environment=environment)
    assert result.success

    assert _load_table(result.context, sum_table_arg) == [(1, 2, 3), (3, 4, 7)]


def test_with_from_through_specifying_all_solids():
    pipeline = create_multi_input_pipeline()

    first_sum_table = 'first_sum_table'
    first_mult_table = 'first_mult_table'
    first_sum_mult_table = 'first_sum_mult_table'

    environment = config.Environment(
        solids={
            'sum_table':
            config.Solid({
                'sum_table': first_sum_table,
            }),
            'mult_table':
            config.Solid({
                'mult_table': first_mult_table,
            }),
            'sum_mult_table':
            config.Solid(
                {
                    'sum_table': first_sum_table,
                    'mult_table': first_mult_table,
                    'sum_mult_table': first_sum_mult_table,
                }
            ),
        },
    )

    pipeline_result = execute_pipeline(pipeline, environment=environment)
    assert len(pipeline_result.result_list) == 3
    assert _load_table(pipeline_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]
    assert _load_table(pipeline_result.context, first_mult_table) == [(1, 2, 2), (3, 4, 12)]
    assert _load_table(pipeline_result.context, first_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]


def test_multi_input_partial_execution():
    pipeline = create_multi_input_pipeline()

    first_sum_table = 'first_sum_table'
    first_mult_table = 'first_mult_table'
    first_sum_mult_table = 'first_sum_mult_table'

    environment = config.Environment(
        solids={
            'sum_table':
            config.Solid({
                'sum_table': first_sum_table
            }),
            'mult_table':
            config.Solid({
                'mult_table': first_mult_table,
            }),
            'sum_mult_table':
            config.Solid(
                {
                    'sum_table': first_sum_table,
                    'mult_table': first_mult_table,
                    'sum_mult_table': first_sum_mult_table,
                }
            ),
        },
    )

    first_pipeline_result = execute_pipeline(pipeline, environment=environment)

    assert first_pipeline_result.success
    assert len(first_pipeline_result.result_list) == 3
    assert _load_table(first_pipeline_result.context, first_sum_table) == [(1, 2, 3), (3, 4, 7)]
    assert _load_table(first_pipeline_result.context, first_mult_table) == [(1, 2, 2), (3, 4, 12)]
    assert _load_table(first_pipeline_result.context,
                       first_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]

    return
    # FIXME: need better API for partial pipeline execution

    # second_sum_mult_table = 'second_sum_mult_table'

    # environment_two = config.Environment(
    #     sources={
    #         'sum_mult_table': {
    #             'sum_table': table_name_source(first_sum_table),
    #             'mult_table': table_name_source(first_mult_table),
    #             'sum_mult_table': table_name_source(second_sum_mult_table)
    #         },
    #     },
    #     execution=config.Execution.single_solid('sum_mult_table')
    # )

    # assert second_pipeline_result.success
    # assert len(second_pipeline_result.result_list) == 1
    # assert _load_table(second_pipeline_result.context,
    #                    second_sum_mult_table) == [(1, 3, 2), (3, 7, 12)]


def create_multi_input_pipeline():
    sum_sql_template = '''CREATE TABLE {{sum_table}} AS
        SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    mult_sql_template = '''CREATE TABLE {{mult_table}} AS
        SELECT num1, num2, num1 * num2 as mult FROM num_table'''

    sum_mult_join_template = '''CREATE TABLE {{sum_mult_table}} AS
        SELECT {{sum_table}}.num1, sum, mult FROM {{sum_table}}
        INNER JOIN {{mult_table}} ON {{sum_table}}.num1 = {{mult_table}}.num1'''

    sum_solid = create_templated_sql_transform_solid(
        name='sum_table',
        sql=sum_sql_template,
        table_arguments=['sum_table'],
    )

    mult_solid = create_templated_sql_transform_solid(
        name='mult_table',
        sql=mult_sql_template,
        table_arguments=['mult_table'],
    )

    sum_mult_solid = create_templated_sql_transform_solid(
        name='sum_mult_table',
        sql=sum_mult_join_template,
        table_arguments=['sum_table', 'mult_table', 'sum_mult_table'],
        dependant_solids=[sum_solid, mult_solid],
    )

    pipeline = pipeline_test_def(
        solids=[sum_solid, mult_solid, sum_mult_solid],
        context=in_mem_context(),
        dependencies={
            sum_mult_solid.name: {
                sum_solid.name: DependencyDefinition(sum_solid.name),
                mult_solid.name: DependencyDefinition(mult_solid.name),
            }
        },
    )
    return pipeline


def test_jinja():
    templated_sql = '''SELECT * FROM {{some_table.name}}'''

    sql = _render_template_string(templated_sql, {'some_table': {'name': 'fill_me_in'}})

    assert sql == '''SELECT * FROM fill_me_in'''



import os
import dagster
import dagster.sqlalchemy as dagster_sa

from dagster import (
    DependencyDefinition,
    InputDefinition,
)


def _get_sql_script_path(name):
    return os.path.join(os.path.dirname(__file__), 'sql_files', '{name}.sql'.format(name=name))


def _get_project_solid(name, inputs=None):
    return dagster_sa.sql_file_solid(_get_sql_script_path(name), inputs=inputs)


def define_full_pipeline():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[InputDefinition(create_all_tables_solids.name)],
    )

    insert_into_sum_table_solid = _get_project_solid(
        'insert_into_sum_table',
        inputs=[InputDefinition(populate_num_table_solid.name)],
    )

    insert_into_sum_sq_table_solid = _get_project_solid(
        'insert_into_sum_sq_table',
        inputs=[InputDefinition(insert_into_sum_table_solid.name)],
    )

    return dagster.PipelineDefinition(
        name='full_pipeline',
        description='Runs entire pipeline, both setup and running the transform',
        solids=[
            create_all_tables_solids,
            populate_num_table_solid,
            insert_into_sum_table_solid,
            insert_into_sum_sq_table_solid,
        ],
        dependencies={
            populate_num_table_solid.name: {
                create_all_tables_solids.name: DependencyDefinition(create_all_tables_solids.name)
            },
            insert_into_sum_table_solid.name: {
                populate_num_table_solid.name: DependencyDefinition(populate_num_table_solid.name),
            },
            insert_into_sum_sq_table_solid.name: {
                insert_into_sum_table_solid.name: DependencyDefinition(insert_into_sum_table_solid),
            }
        },
    )


def define_truncate_pipeline():
    truncate_solid = _get_project_solid('truncate_all_derived_tables')
    return dagster.PipelineDefinition(
        name='truncate_all_derived_tables',
        description=
        'Truncates all tables that are populated by the pipeline. Preserves source tables',
        solids=[truncate_solid]
    )


def define_rerun_pipeline():
    insert_into_sum_table_solid = _get_project_solid(
        'insert_into_sum_table',
        inputs=None,
    )

    insert_into_sum_sq_table_solid = _get_project_solid(
        'insert_into_sum_sq_table', inputs=[InputDefinition(insert_into_sum_table_solid.name)]
    )

    return dagster.PipelineDefinition(
        name='rerun_pipeline',
        description=
        'Rerun the pipeline, populating the the derived tables. Assumes pipeline is setup',
        solids=[
            insert_into_sum_table_solid,
            insert_into_sum_sq_table_solid,
        ],
        dependencies={
            insert_into_sum_sq_table_solid.name: {
                insert_into_sum_table_solid.name:
                DependencyDefinition(insert_into_sum_table_solid.name),
            }
        }
    )


def define_setup_pipeline():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[InputDefinition(create_all_tables_solids.name)],
    )

    return dagster.PipelineDefinition(
        name='setup_pipeline',
        description='Creates all tables and then populates source table',
        solids=[
            create_all_tables_solids,
            populate_num_table_solid,
        ],
        dependencies={
            populate_num_table_solid.name: {
                create_all_tables_solids.name: DependencyDefinition(create_all_tables_solids.name),
            }
        }
    )



import pytest

from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    ExecutionContext,
    InputDefinition,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    Result,
    SolidDefinition,
    check,
    execute_pipeline,
)

from dagster.core.test_utils import (
    execute_single_solid,
    single_output_transform,
)
from dagster.core.errors import DagsterUserCodeExecutionError


def silencing_default_context():
    return {'default': PipelineContextDefinition(context_fn=lambda *_args: ExecutionContext())}


def silencing_pipeline(solids, dependencies=None):
    return PipelineDefinition(
        solids=solids,
        dependencies=dependencies,
        context_definitions=silencing_default_context(),
    )


def create_root_success_solid(name):
    def root_transform(_context, _args):
        passed_rows = []
        passed_rows.append({name: 'transform_called'})
        return passed_rows

    return single_output_transform(
        name=name,
        inputs=[],
        transform_fn=root_transform,
        output=OutputDefinition(),
    )


def create_root_transform_failure_solid(name):
    def failed_transform(**_kwargs):
        raise Exception('Transform failed')

    return single_output_transform(
        name=name,
        inputs=[],
        transform_fn=failed_transform,
        output=OutputDefinition(),
    )


def test_transform_failure_pipeline():
    pipeline = silencing_pipeline(solids=[create_root_transform_failure_solid('failing')])
    pipeline_result = execute_pipeline(pipeline, throw_on_error=False)

    assert not pipeline_result.success

    result_list = pipeline_result.result_list

    assert len(result_list) == 1
    assert not result_list[0].success
    assert result_list[0].dagster_error


def test_failure_midstream():
    solid_a = create_root_success_solid('A')
    solid_b = create_root_success_solid('B')

    def transform_fn(_context, inputs):
        check.failed('user error')
        return [inputs['A'], inputs['B'], {'C': 'transform_called'}]

    solid_c = single_output_transform(
        name='C',
        inputs=[InputDefinition(name='A'), InputDefinition(name='B')],
        transform_fn=transform_fn,
        output=OutputDefinition(),
    )

    pipeline = silencing_pipeline(
        solids=[solid_a, solid_b, solid_c],
        dependencies={
            'C': {
                'A': DependencyDefinition(solid_a.name),
                'B': DependencyDefinition(solid_b.name),
            }
        }
    )
    pipeline_result = execute_pipeline(
        pipeline,
        throw_on_error=False,
    )

    result_list = pipeline_result.result_list

    assert result_list[0].success
    assert result_list[1].success
    assert not result_list[2].success
    assert isinstance(result_list[2].dagster_error, DagsterUserCodeExecutionError)


def test_do_not_yield_result():
    solid_inst = SolidDefinition(
        name='do_not_yield_result',
        inputs=[],
        outputs=[OutputDefinition()],
        transform_fn=lambda *_args, **_kwargs: Result('foo')
    )

    with pytest.raises(
        DagsterInvariantViolationError,
        message='Tranform for solid do_not_yield_result return a Result',
    ):
        execute_single_solid(ExecutionContext(), solid_inst)


def test_yield_non_result():
    def _tn(*_args, **_kwargs):
        yield 'foo'

    solid_inst = SolidDefinition(
        name='yield_wrong_thing',
        inputs=[],
        outputs=[OutputDefinition()],
        transform_fn=_tn,
    )

    with pytest.raises(
        DagsterInvariantViolationError,
        message="Tranform for solid yield_wrong_thing yielded 'foo'",
    ):
        execute_single_solid(ExecutionContext(), solid_inst)


def test_single_transform_returning_result():
    solid_inst = single_output_transform(
        'test_return_result',
        inputs=[],
        transform_fn=lambda *_args, **_kwargs: Result(None),
        output=OutputDefinition()
    )

    with pytest.raises(DagsterInvariantViolationError):
        execute_single_solid(ExecutionContext(), solid_inst)



from dagster import (
    ExecutionContext,
    PipelineContextDefinition,
    PipelineDefinition,
    config,
    lambda_solid,
)

from dagster.core.execution import (
    ComputeNodeExecutionInfo,
)

from dagster.core.definitions import ExecutionGraph

from dagster.core.compute_nodes import create_compute_node_graph


def silencing_default_context():
    return {'default': PipelineContextDefinition(context_fn=lambda *_args: ExecutionContext(), )}


@lambda_solid
def noop():
    return 'foo'


def silencing_pipeline(solids):
    return PipelineDefinition(solids=solids, context_definitions=silencing_default_context())


def test_compute_noop_node():
    pipeline = silencing_pipeline(solids=[
        noop,
    ])

    environment = config.Environment()

    execution_graph = ExecutionGraph.from_pipeline(pipeline)
    compute_node_graph = create_compute_node_graph(
        ComputeNodeExecutionInfo(
            ExecutionContext(),
            execution_graph,
            environment,
        ),
    )

    assert len(compute_node_graph.nodes) == 1

    outputs = list(compute_node_graph.nodes[0].execute(ExecutionContext(), {}))

    assert outputs[0].success_data.value == 'foo'



from collections import defaultdict

from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    InputDefinition,
    PipelineDefinition,
    SolidInstance,
    config,
    execute_pipeline,
    lambda_solid,
    solid,
    types,
)


def test_aliased_solids():
    @lambda_solid()
    def first():
        return ['first']

    @lambda_solid(inputs=[InputDefinition(name="prev")])
    def not_first(prev):
        return prev + ['not_first']

    pipeline = PipelineDefinition(
        solids=[first, not_first],
        dependencies={
            'not_first': {
                'prev': DependencyDefinition('first'),
            },
            SolidInstance('not_first', alias='second'): {
                'prev': DependencyDefinition('not_first'),
            },
            SolidInstance('not_first', alias='third'): {
                'prev': DependencyDefinition('second'),
            },
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success
    solid_result = result.result_for_solid('third')
    assert solid_result.transformed_value() == ['first', 'not_first', 'not_first', 'not_first']


def test_only_aliased_solids():
    @lambda_solid()
    def first():
        return ['first']

    @lambda_solid(inputs=[InputDefinition(name="prev")])
    def not_first(prev):
        return prev + ['not_first']

    pipeline = PipelineDefinition(
        solids=[first, not_first],
        dependencies={
            SolidInstance('first', alias='the_root'): {},
            SolidInstance('not_first', alias='the_consequence'): {
                'prev': DependencyDefinition('the_root'),
            },
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success
    solid_result = result.result_for_solid('the_consequence')
    assert solid_result.transformed_value() == ['first', 'not_first']


def test_aliased_configs():
    @solid(
        inputs=[],
        config_def=ConfigDefinition(types.Int),
    )
    def load_constant(info):
        return info.config

    pipeline = PipelineDefinition(
        solids=[load_constant],
        dependencies={
            SolidInstance(load_constant.name, 'load_a'): {},
            SolidInstance(load_constant.name, 'load_b'): {},
        }
    )

    result = execute_pipeline(
        pipeline,
        config.Environment(solids={
            'load_a': config.Solid(2),
            'load_b': config.Solid(3),
        })
    )

    assert result.success
    assert result.result_for_solid('load_a').transformed_value() == 2
    assert result.result_for_solid('load_b').transformed_value() == 3


def test_aliased_solids_context():
    record = defaultdict(set)

    @solid
    def log_things(info):
        solid_value = info.context.get_context_value('solid')
        solid_def_value = info.context.get_context_value('solid_definition')
        record[solid_def_value].add(solid_value)

    pipeline = PipelineDefinition(
        solids=[log_things],
        dependencies={
            SolidInstance('log_things', 'log_a'): {},
            SolidInstance('log_things', 'log_b'): {},
        }
    )

    result = execute_pipeline(pipeline)
    assert result.success

    assert record == {'log_things': set(['log_a', 'log_b'])}



from dagster.core.execution import (
    PipelineExecutionResult,
    SolidExecutionResult,
    execute_pipeline,
    execute_pipeline_iterator,
)

from dagster.core.execution_context import (
    ExecutionContext,
)

from dagster.core.definitions import (
    ConfigDefinition,
    ContextCreationExecutionInfo,
    DependencyDefinition,
    ExpectationDefinition,
    ExpectationExecutionInfo,
    ExpectationResult,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    RepositoryDefinition,
    Result,
    SolidDefinition,
    SolidInstance,
    TransformExecutionInfo,
)

from dagster.core.decorators import (
    MultipleResults,
    lambda_solid,
    solid,
)

from dagster.core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterTypeError,
    DagsterUserCodeExecutionError,
    DagsterExpectationFailedError,
    DagsterEvaluateValueError,
)

from dagster.core.utility_solids import define_stub_solid

import dagster.config as config
import dagster.core.types as types

from .version import __version__

__all__ = [
    # Definition
    'ConfigDefinition',
    'DependencyDefinition',
    'ExpectationDefinition',
    'ExpectationResult',
    'Field',
    'InputDefinition',
    'OutputDefinition',
    'PipelineContextDefinition',
    'PipelineDefinition',
    'RepositoryDefinition',
    'SolidDefinition',
    'Result',
    'SolidInstance',

    # Infos
    'ContextCreationExecutionInfo',
    'ExpectationExecutionInfo',
    'TransformExecutionInfo',

    # Decorators
    'lambda_solid',
    'solid',
    'MultipleResults',

    # Execution
    'execute_pipeline_iterator',
    'execute_pipeline',
    'ExecutionContext',
    'PipelineExecutionResult',
    'SolidExecutionResult',

    # Errors
    'DagsterInvalidDefinitionError',
    'DagsterInvariantViolationError',
    'DagsterTypeError',
    'DagsterUserCodeExecutionError',
    'DagsterExpectationFailedError',
    'DagsterEvaluateValueError',

    # utility_solids
    'define_stub_solid',

    # config
    'config',

    # types
    'types',
]



from collections import namedtuple
from functools import wraps
import inspect

from .definitions import (
    ConfigDefinition,
    DagsterInvalidDefinitionError,
    InputDefinition,
    OutputDefinition,
    Result,
    SolidDefinition,
    check,
)

if hasattr(inspect, 'signature'):
    funcsigs = inspect
else:
    import funcsigs

# Error messages are long
# pylint: disable=C0301


class MultipleResults(namedtuple('_MultipleResults', 'results')):
    '''A shortcut to output multiple results.

    Attributes:
      results (list[Result]): list of :py:class:`Result`

    Example:

    .. code-block:: python

        @solid(outputs=[
            OutputDefinition(name='foo'),
            OutputDefinition(name='bar'),
        ])
        def my_solid():
            return MultipleResults(
                Result('Barb', 'foo'),
                Result('Glarb', 'bar'),
            )


        @solid(outputs=[
            OutputDefinition(name='foo'),
            OutputDefinition(name='bar'),
        ])
        def my_solid_from_dict():
            return MultipleResults.from_dict({
              'foo': 'Barb',
              'bar': 'Glarb',
            })
    '''

    def __new__(cls, *results):
        return super(MultipleResults, cls).__new__(
            cls,
            # XXX(freiksenet): should check.list_param accept tuples from rest?
            check.opt_list_param(list(results), 'results', Result),
        )

    @staticmethod
    def from_dict(result_dict):
        '''Create MultipleResults object from a dictionary. Keys become result names'''
        check.dict_param(result_dict, 'result_dict', key_type=str)
        results = []
        for name, value in result_dict.items():
            results.append(Result(value, name))
        return MultipleResults(*results)


class _LambdaSolid(object):
    def __init__(
        self,
        name=None,
        inputs=None,
        output=None,
        description=None,
    ):
        self.name = check.opt_str_param(name, 'name')
        self.input_defs = check.opt_list_param(inputs, 'inputs', InputDefinition)
        self.output_def = check.inst_param(output, 'output', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        _validate_transform_fn(self.name, fn, self.input_defs)
        transform_fn = _create_lambda_solid_transform_wrapper(fn, self.input_defs, self.output_def)
        return SolidDefinition(
            name=self.name,
            inputs=self.input_defs,
            outputs=[self.output_def],
            transform_fn=transform_fn,
            description=self.description,
        )


class _Solid(object):
    def __init__(
        self,
        name=None,
        inputs=None,
        outputs=None,
        description=None,
        config_def=None,
    ):
        self.name = check.opt_str_param(name, 'name')
        self.input_defs = check.opt_list_param(inputs, 'inputs', InputDefinition)
        outputs = outputs or [OutputDefinition()]
        self.outputs = check.list_param(outputs, 'outputs', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')
        self.config_def = check.opt_inst_param(config_def, 'config_def', ConfigDefinition)

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        _validate_transform_fn(self.name, fn, self.input_defs, ['info'])
        transform_fn = _create_solid_transform_wrapper(fn, self.input_defs, self.outputs)
        return SolidDefinition(
            name=self.name,
            inputs=self.input_defs,
            outputs=self.outputs,
            transform_fn=transform_fn,
            config_def=self.config_def,
            description=self.description,
        )


def lambda_solid(
    name=None,
    inputs=None,
    output=None,
    description=None,
):
    '''(decorator) Create a simple solid

    This shortcut allows the creation of simple solids that do not require
    configuration and whose implementations do not require a context.

    Lambda solids take inputs an and produce an output. The body of the function
    should return a single value.

    Args:
        name (str): Name of solid
        inputs (List[InputDefinition]): List of inputs
        output (OutputDefinition): The output of the solid. Defaults to OutputDefinition()
        description (str): Solid description

    .. code-block:: python

        @lambda_solid
        def hello_world():
            return 'hello'

        @lambda_solid(inputs=[InputDefinition(name="foo")])
        def hello_world(foo):
            return foo

    '''
    output = output or OutputDefinition()

    if callable(name):
        check.invariant(inputs is None)
        check.invariant(description is None)
        return _LambdaSolid(output=output)(name)

    return _LambdaSolid(
        name=name,
        inputs=inputs,
        output=output,
        description=description,
    )


def solid(
    name=None,
    inputs=None,
    outputs=None,
    config_def=None,
    description=None,
):
    '''(decorator) Create a solid with specified parameters.

    This shortcut simplifies core solid API by exploding arguments into kwargs of the
    transform function and omitting additional parameters when they are not needed.
    Parameters are otherwise as per :py:class:`SolidDefinition`.

    Decorated function is the transform function itself. Instead of having to yield
    result objects, transform support multiple simpler output types.

    1. Return a value. This is returned as a :py:class:`Result` for a single output solid.
    2. Return a :py:class:`Result`. Works like yielding result.
    3. Return a :py:class:`MultipleResults`. Works like yielding several results for
       multiple outputs. Useful for solids that have multiple outputs.
    4. Yield :py:class:`Result`. Same as default transform behaviour.

    Args:
        name (str): Name of solid
        inputs (List[InputDefinition]): List of inputs
        outputs (List[OutputDefinition]): List of outputs
        config_def (ConfigDefinition):
            The configuration for this solid.
        description (str): Description of this solid.

    Examples:

    .. code-block:: python

        @solid
        def hello_world(info):
            print('hello')

        @solid()
        def hello_world(info):
            print('hello')

        @solid(outputs=[OutputDefinition()])
        def hello_world(info):
            return {'foo': 'bar'}

        @solid(outputs=[OutputDefinition()])
        def hello_world(info):
            return Result(value={'foo': 'bar'})

        @solid(outputs=[OutputDefinition()])
        def hello_world(info):
            yield Result(value={'foo': 'bar'})

        @solid(outputs=[
            OutputDefinition(name="left"),
            OutputDefinition(name="right"),
        ])
        def hello_world(info):
            return MultipleResults.from_dict({
                'left': {'foo': 'left'},
                'right': {'foo': 'right'},
            })

        @solid(
            inputs=[InputDefinition(name="foo")],
            outputs=[OutputDefinition()]
        )
        def hello_world(info, foo):
            return foo

        @solid(
            inputs=[InputDefinition(name="foo")],
            outputs=[OutputDefinition()],
        )
        def hello_world(info, foo):
            info.context.info('log something')
            return foo

        @solid(
            inputs=[InputDefinition(name="foo")],
            outputs=[OutputDefinition()],
            config_def=ConfigDefinition(types.ConfigDictionary({'str_value' : Field(types.String)})),
        )
        def hello_world(info, foo):
            # info.config is a dictionary with 'str_value' key
            return foo + info.config['str_value']

    '''
    if callable(name):
        check.invariant(inputs is None)
        check.invariant(outputs is None)
        check.invariant(description is None)
        check.invariant(config_def is None)
        return _Solid()(name)

    return _Solid(
        name=name,
        inputs=inputs,
        outputs=outputs,
        config_def=config_def,
        description=description,
    )


def _create_lambda_solid_transform_wrapper(fn, input_defs, output_def):
    check.callable_param(fn, 'fn')
    check.list_param(input_defs, 'input_defs', of_type=InputDefinition)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    input_names = [input_def.name for input_def in input_defs]

    @wraps(fn)
    def transform(_info, inputs):
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = inputs[input_name]

        result = fn(**kwargs)
        yield Result(value=result, output_name=output_def.name)

    return transform


def _create_solid_transform_wrapper(fn, input_defs, output_defs):
    check.callable_param(fn, 'fn')
    check.list_param(input_defs, 'input_defs', of_type=InputDefinition)
    check.list_param(output_defs, 'output_defs', of_type=OutputDefinition)

    input_names = [input_def.name for input_def in input_defs]

    @wraps(fn)
    def transform(info, inputs):
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = inputs[input_name]

        result = fn(info, **kwargs)

        if inspect.isgenerator(result):
            for item in result:
                yield item
        else:
            if isinstance(result, Result):
                yield result
            elif isinstance(result, MultipleResults):
                for item in result.results:
                    yield item
            elif len(output_defs) == 1:
                yield Result(value=result, output_name=output_defs[0].name)
            elif result is not None:
                # XXX(freiksenet)
                raise Exception('Output for a solid without an output.')

    return transform


class FunctionValidationError(Exception):
    TYPES = {
        'vararg': 1,
        'missing_name': 2,
        'missing_positional': 3,
        'extra': 4,
    }

    def __init__(self, error_type, param=None, missing_names=None, **kwargs):
        super(FunctionValidationError, self).__init__(**kwargs)
        self.error_type = error_type
        self.param = param
        self.missing_names = missing_names


def _validate_transform_fn(solid_name, transform_fn, inputs, expected_positionals=None):
    check.str_param(solid_name, 'solid_name')
    check.callable_param(transform_fn, 'transform_fn')
    check.list_param(inputs, 'inputs', of_type=InputDefinition)
    expected_positionals = check.opt_list_param(
        expected_positionals,
        'expected_positionals',
        of_type=str,
    )

    names = set(inp.name for inp in inputs)
    # Currently being super strict about naming. Might be a good idea to relax. Starting strict.
    try:
        _validate_decorated_fn(transform_fn, names, expected_positionals)
    except FunctionValidationError as e:
        if e.error_type == FunctionValidationError.TYPES['vararg']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function has positional vararg parameter '{e.param}'. Transform functions should only have keyword arguments that match input names and a first positional parameter named 'info'.".
                format(solid_name=solid_name, e=e)
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_name']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function has parameter '{e.param}' that is not one of the solid inputs. Transform functions should only have keyword arguments that match input names and a first positional parameter named 'info'.".
                format(solid_name=solid_name, e=e)
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_positional']:
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function do not have required positional parameter '{e.param}'. Transform functions should only have keyword arguments that match input names and a first positional parameter named 'info'.".
                format(solid_name=solid_name, e=e)
            )
        elif e.error_type == FunctionValidationError.TYPES['extra']:
            undeclared_inputs_printed = ", '".join(e.missing_names)
            raise DagsterInvalidDefinitionError(
                "solid '{solid_name}' transform function do not have parameter(s) '{undeclared_inputs_printed}', which are in solid's inputs. Transform functions should only have keyword arguments that match input names and a first positional parameter named 'info'.".
                format(solid_name=solid_name, undeclared_inputs_printed=undeclared_inputs_printed)
            )
        else:
            raise e


def _validate_decorated_fn(fn, names, expected_positionals):
    used_inputs = set()
    has_kwargs = False

    signature = funcsigs.signature(fn)
    params = list(signature.parameters.values())

    expected_positional_params = params[0:len(expected_positionals)]
    other_params = params[len(expected_positionals):]

    for expected, actual in zip(expected_positionals, expected_positional_params):
        possible_names = [
            '_',
            expected,
            '_{expected}'.format(expected=expected),
            '{expected}_'.format(expected=expected),
        ]
        if (
            actual.kind not in [
                funcsigs.Parameter.POSITIONAL_OR_KEYWORD, funcsigs.Parameter.POSITIONAL_ONLY
            ]
        ) or (actual.name not in possible_names):
            raise FunctionValidationError(
                FunctionValidationError.TYPES['missing_positional'], param=expected
            )

    for param in other_params:
        if param.kind == funcsigs.Parameter.VAR_KEYWORD:
            has_kwargs = True
        elif param.kind == funcsigs.Parameter.VAR_POSITIONAL:
            raise FunctionValidationError(error_type=FunctionValidationError.TYPES['vararg'])

        else:
            if param.name not in names:
                raise FunctionValidationError(
                    FunctionValidationError.TYPES['missing_name'], param=param.name
                )
            else:
                used_inputs.add(param.name)

    undeclared_inputs = names - used_inputs
    if not has_kwargs and undeclared_inputs:
        raise FunctionValidationError(
            FunctionValidationError.TYPES['extra'], missing_names=undeclared_inputs
        )



import pandas as pd

from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    config,
    execute_pipeline,
    lambda_solid,
    PipelineDefinition,
)

from dagster.core.test_utils import single_output_transform

from dagster.utils import script_relative_path
from dagster.utils.test import get_temp_file_name

import dagster.pandas as dagster_pd


def create_num_csv_environment():
    return config.Environment(
        solids={'load_csv': config.Solid({
            'path': script_relative_path('num.csv')
        })},
    )


def test_hello_world_with_dataframe_fns():
    hello_world = create_definition_based_solid()
    run_hello_world(hello_world)


def run_hello_world(hello_world):
    assert len(hello_world.input_defs) == 1

    pipeline = PipelineDefinition(
        solids=[
            dagster_pd.load_csv_solid('load_csv'),
            hello_world,
        ],
        dependencies={
            'hello_world': {
                'num_csv': DependencyDefinition('load_csv'),
            },
        }
    )

    pipeline_result = execute_pipeline(
        pipeline,
        environment=create_num_csv_environment(),
    )

    result = pipeline_result.result_for_solid('hello_world')

    assert result.success

    assert result.transformed_value().to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }

    pipeline_two = PipelineDefinition(
        solids=[
            dagster_pd.load_csv_solid('load_csv'),
            hello_world,
            dagster_pd.to_csv_solid('to_csv'),
        ],
        dependencies={
            'hello_world': {
                'num_csv': DependencyDefinition('load_csv'),
            },
            'to_csv': {
                'df': DependencyDefinition('hello_world'),
            }
        }
    )

    with get_temp_file_name() as temp_file_name:
        environment = config.Environment(
            solids={
                'load_csv': config.Solid({
                    'path': script_relative_path('num.csv'),
                }),
                'to_csv': config.Solid({
                    'path': temp_file_name,
                })
            },
        )
        pipeline_result = execute_pipeline(
            pipeline_two,
            environment,
        )

        output_result = pipeline_result.result_for_solid('hello_world')

        assert output_result.success

        assert pd.read_csv(temp_file_name).to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
        }


def create_definition_based_solid():
    table_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    def transform_fn(_context, inputs):
        num_csv = inputs['num_csv']
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    # supports CSV and PARQUET by default
    hello_world = single_output_transform(
        name='hello_world',
        inputs=[table_input],
        transform_fn=transform_fn,
        output=OutputDefinition(dagster_pd.DataFrame)
    )
    return hello_world


def create_decorator_based_solid():
    @lambda_solid(
        inputs=[InputDefinition('num_csv', dagster_pd.DataFrame)],
        output=OutputDefinition(dagster_pd.DataFrame),
    )
    def hello_world(num_csv):
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    return hello_world


def test_hello_world_decorator_style():
    hello_world = create_decorator_based_solid()
    run_hello_world(hello_world)



import sqlalchemy as sa

from dagster import (
    DependencyDefinition,
    InputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    check,
    config,
    execute_pipeline,
)

from dagster.core.utility_solids import define_stub_solid

from dagster.sqlalchemy.subquery_builder_experimental import (
    create_sql_solid,
    DagsterSqlTableExpression,
    define_create_table_solid,
)
from .math_test_db import in_mem_context


def pipeline_test_def(solids, context, dependencies):
    return PipelineDefinition(
        solids=solids,
        context_definitions={
            'default': PipelineContextDefinition(context_fn=lambda *_args: context),
        },
        dependencies=dependencies,
    )


def create_num_table(engine):
    metadata = sa.MetaData(engine)

    table = sa.Table(
        'num_table',
        metadata,
        sa.Column('num1', sa.Integer),
        sa.Column('num2', sa.Integer),
    )

    table.create()

    conn = engine.connect()

    conn.execute('''INSERT INTO num_table VALUES(1, 2)''')
    conn.execute('''INSERT INTO num_table VALUES(3, 4)''')


def test_sql_sum_solid():
    expr_solid = define_stub_solid('expr', DagsterSqlTableExpression('num_table'))

    sum_table_solid = create_sum_table_solid()

    create_sum_table = define_create_table_solid('create_sum_table_solid')

    environment = config.Environment(
        solids={create_sum_table.name: config.Solid({
            'table_name': 'sum_table',
        })}
    )

    pipeline = pipeline_test_def(
        solids=[expr_solid, sum_table_solid, create_sum_table],
        context=in_mem_context(),
        dependencies={
            'sum_table': {
                'num_table': DependencyDefinition('expr')
            },
            create_sum_table.name: {
                'expr': DependencyDefinition('sum_table'),
            }
        },
    )

    pipeline_result = execute_pipeline(pipeline, environment)
    assert pipeline_result.success

    result = pipeline_result.result_for_solid(sum_table_solid.name)

    assert result.success

    results = result.context.resources.sa.engine.connect().execute('SELECT * FROM sum_table'
                                                                   ).fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]


def create_sum_table_solid():
    return create_sql_solid(
        name='sum_table',
        inputs=[InputDefinition('num_table')],
        sql_text='SELECT num1, num2, num1 + num2 as sum FROM {num_table}',
    )


def create_sum_sq_pipeline(context, expr, extra_solids=None, extra_deps=None):
    check.inst_param(expr, 'expr', DagsterSqlTableExpression)

    expr_solid = define_stub_solid('expr', expr)

    sum_solid = create_sum_table_solid()

    sum_sq_solid = create_sql_solid(
        name='sum_sq_table',
        inputs=[InputDefinition(sum_solid.name)],
        sql_text='SELECT num1, num2, sum, sum * sum as sum_sq from {sum_table}',
    )

    dependencies = {
        sum_solid.name: {
            'num_table': DependencyDefinition('expr'),
        },
        sum_sq_solid.name: {
            sum_solid.name: DependencyDefinition(sum_solid.name),
        }
    }
    if extra_deps:
        dependencies.update(extra_deps)

    return pipeline_test_def(
        solids=[expr_solid, sum_solid, sum_sq_solid] + (extra_solids if extra_solids else []),
        context=context,
        dependencies=dependencies
    )


def test_execute_sql_sum_sq_solid():
    pipeline = create_sum_sq_pipeline(in_mem_context(), DagsterSqlTableExpression('num_table'))

    pipeline_result = execute_pipeline(pipeline)

    assert pipeline_result.success

    result_list = pipeline_result.result_list

    sum_table_sql_text = result_list[1].transformed_value().query_text
    assert sum_table_sql_text == 'SELECT num1, num2, num1 + num2 as sum FROM num_table'

    sum_sq_table_sql_text = result_list[2].transformed_value().query_text
    assert sum_sq_table_sql_text == 'SELECT num1, num2, sum, sum * sum as sum_sq from ' + \
            '(SELECT num1, num2, num1 + num2 as sum FROM num_table)'


def test_output_sql_sum_sq_solid():
    create_sum_sq_table = define_create_table_solid('create_sum_sq_table')

    pipeline = create_sum_sq_pipeline(
        in_mem_context(), DagsterSqlTableExpression('num_table'), [create_sum_sq_table],
        {create_sum_sq_table.name: {
            'expr': DependencyDefinition('sum_sq_table')
        }}
    )

    environment = config.Environment(
        solids={'create_sum_sq_table': config.Solid({
            'table_name': 'sum_sq_table'
        })},
    )

    pipeline_result = execute_pipeline(pipeline=pipeline, environment=environment)

    assert pipeline_result.success

    result_list = pipeline_result.result_list

    assert len(result_list) == 3
    engine = pipeline_result.context.resources.sa.engine
    result_list = engine.connect().execute('SELECT * FROM sum_sq_table').fetchall()
    assert result_list == [(1, 2, 3, 9), (3, 4, 7, 49)]



from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import *  # pylint: disable=W0622,W0401

import pandas as pd

from dagster import (
    ConfigDefinition,
    Field,
    ExecutionContext,
    InputDefinition,
    OutputDefinition,
    Result,
    SolidDefinition,
    check,
    types,
)


def _create_dataframe_type():
    return types.PythonObjectType(
        name='PandasDataFrame',
        python_type=pd.DataFrame,
        description='''Two-dimensional size-mutable, potentially heterogeneous
tabular data structure with labeled axes (rows and columns). See http://pandas.pydata.org/''',
    )


DataFrame = _create_dataframe_type()

LoadDataFrameConfigDict = types.ConfigDictionary(
    'LoadDataFrameConfigDict',
    {
        'path': Field(types.Path),
    },
)

WriteDataFrameConfigDict = types.ConfigDictionary(
    'WriteDataFrameConfigDict',
    {
        'path': Field(types.Path),
    },
)


def load_csv_solid(name):
    check.str_param(name, 'name')

    def _t_fn(info, _inputs):
        yield Result(pd.read_csv(info.config['path']))

    return SolidDefinition(
        name=name,
        inputs=[],
        outputs=[OutputDefinition(DataFrame)],
        transform_fn=_t_fn,
        config_def=ConfigDefinition(LoadDataFrameConfigDict),
    )


def to_csv_solid(name):
    def _t_fn(info, inputs):
        inputs['df'].to_csv(info.config['path'], index=False)

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition('df', DataFrame)],
        outputs=[],
        config_def=ConfigDefinition(WriteDataFrameConfigDict),
        transform_fn=_t_fn,
    )


def to_parquet_solid(name):
    def _t_fn(info, inputs):
        inputs['df'].to_parquet(info.config['path'])

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition('df', DataFrame)],
        outputs=[],
        config_def=ConfigDefinition(WriteDataFrameConfigDict),
        transform_fn=_t_fn,
    )
